# -*- coding: utf-8 -*-
"""
.. module:: pytfa
   :platform: Unix, Windows
   :synopsis: Thermodynamics-based Flux Analysis

.. moduleauthor:: pyTFA team

Variability analysis

"""
from tqdm import tqdm
import pandas as pd
import numpy as np
from numpy.linalg import norm
from optlang.interface import Constraint
from pytfa.optim.variables import GenericVariable,ModelVariable
from pytfa.optim.constraints import GenericConstraint
from warnings import warn

# from ..optim.variables import GenericVariable,ModelVariable

BIGM = 1000

class ChebyshevRadius(ModelVariable):
    """
    Variable representing a Chebyshev radius
    """

    prefix = 'CR_'

def is_inequality(constraint):

    if not isinstance(constraint, Constraint):
        the_cons = constraint.constraint
    else:
        the_cons = constraint

    # If one of the bounds is None, it's an inequality
    return the_cons.lb is None or the_cons.ub is None


def chebyshev_center(model, variables, inplace = False, big_m=BIGM,
                     include = list(), exclude=list()):
    """
    Computes the chebyshev center of a problem with respect to given variables,
    including `include' constraints and excluding `exclude' constraints.
    *Warning: Only works with pyTFA variables so far*

    :param model:
    :param variables:
    :param inplace:
    :param big_m:
    :return:
    """

    if not inplace:
        new = model.copy()
        new.optimize()
    else:
        new = model

    vars = get_variables(new, variables)
    include_list = get_cons_var_classes(new, include, type = 'cons')
    exclude_list = get_cons_var_classes(new, exclude, type = 'cons')

    r = chebyshev_transform(model=new,
                            vars=vars,
                            include_list=include_list,
                            exclude_list=exclude_list,
                            big_m=big_m)

    new.objective.direction = 'max'
    new.objective = r.variable
    new.optimize()

    print('Chebyshev Radius: {}'.format(r.variable.primal))
    if r.variable.ub-r.variable.primal <= new.solver.configuration.tolerances.optimality:
        warn('Chebyshev Radius is close to the upper '
             'bound {}. Change the big_m argument to a bigger one.'.format(r.ub))
    var_values = {k.name:k.primal for k in vars}

    return pd.Series(var_values)


def chebyshev_transform(model, vars, include_list=list(), exclude_list=list(),
                        radius_id ='radius',
                        scaling_factor =1,
                        big_m=BIGM):
    """
    Adds a Chebyshev radius variable and edits accordingly the selected
    constraints

    :param model:
    :param vars: variables with respect to which to perform the Chebyshev
        centering. If none is supplied, all of the variables in the equation
        will be considered
    :param include_list:
    :param exclude_list:
    :param radius_id:
    :param big_m:
    :return:
    """
    # 0 - Create the Chebyshev radius variable
    r = model.add_variable(kind=ChebyshevRadius,
                           hook=model,
                           id_=radius_id,
                           lb=0,
                           ub=big_m,
                           queue=False)
    # 1 - Find the inequalities associated with the variables
    # of type a_i*x - b_i <= 0
    # Enumerate the constraints, check which ones are:
    #   - Inequalities
    #   - Containing at least 1 of the given variables
    cons_to_edit = dict()
    for cons in tqdm(model._cons_dict.values(), desc='Finding const.'):
        if type(cons) in exclude_list or type(cons) not in include_list:
            continue
        if not is_inequality(cons):
            continue

        if len(vars) > 0:
            var_intersection = set(cons.expr.free_symbols).intersection(vars)
            if not var_intersection:
                continue
        else:
            var_intersection = cons.expr.free_symbols

        # 2 - For each inequality, find the norm of the vector of coefficients
        # for the variables
        # ||a_i||_2 = sqrt(sum(x**2 for x in coeffs of variables in this eq))
        a_i = {x: cons.expr.coeff(x) for x in var_intersection}
        a_sq = norm(np.array(list(a_i.values()), dtype=float), ord=2)

        cons_to_edit[cons] = a_sq
    # 3 - Replace the constraint bu the same constraint plus the Chebyshev slack
    # a_i*x + ||a_i||_2 * r - b_i <= 0
    for cons, a_sq in tqdm(cons_to_edit.items(), desc='Editing const.'):

        new_expr = cons.expr

        if cons.constraint.lb is None:
            # It's a <= 0 constraint
            new_expr += a_sq * scaling_factor * r
        elif cons.constraint.ub is None:
            # It's a >=0 constraint
            new_expr -= a_sq * scaling_factor * r

        cons.change_expr(new_expr)
    model.logger.info('{} constraints edited with variable {}.'
                      .format(len(cons_to_edit),radius_id))
    # 4 - Optimize
    model.repair()  # Add the queued constraints
    return r


def get_cons_var_classes(model, elements, type):

    if len(elements) == 0:
        return []

    if type.lower().startswith('var'):
        GenericClass = GenericVariable
        model_elements = model._var_kinds
    elif type.lower().startswith('cons'):
        GenericClass = GenericConstraint
        model_elements = model._cons_kinds

    # For safety,
    # Update the variable indices
    model.regenerate_variables()
    model.regenerate_constraints()

    if isinstance(elements[0], str):
        ret = [model_elements[elt][0].__class__ for elt in elements]
    elif issubclass(elements[0], GenericClass):
        ret = elements

    return ret



def get_variables(model, variables):
    if isinstance(variables[0], str):
        # These are var names, we have to retrieve the optlang variables
        vars = [model.variables.get(x) for x in variables]
    elif isinstance(variables[0], GenericVariable):
        # These are pyTFA variables, we have to retrieve the optlang variables
        vars = [model._var_dict[x.name].variable for x in variables]
    return vars




# -*- coding: utf-8 -*-
"""
.. module:: pytfa
   :platform: Unix, Windows
   :synopsis: Thermodynamics-based Flux Analysis

.. moduleauthor:: pyTFA team

Constraints declarations

"""

from ..utils.str import camel2underscores
from .meta import ABCRequirePrefixMeta

###################################################
###                 CONSTRAINTS                 ###
###################################################


class GenericConstraint(metaclass=ABCRequirePrefixMeta):
    """
    Class to represent a generic constraint. The purpose is that the interface
     is instantiated on initialization, to follow the type of interface used
     by the problem, and avoid incompatibilities in optlang

    Attributes:

        :id: Used for DictList comprehension. Usually points back at a
        enzyme or reaction id for ease of linking. Should be unique given
        a constraint type.
        :name: Should be a concatenation of the id and a prefix that is
        specific to the variable type. will be used to address the constraint at
        the solver level, and hence should be unique in the whole cobra_model
        :expr: the expression of the constraint (sympy.Expression subtype)
        :cobra_model: the cobra_model hook.
        :constraint: links directly to the cobra_model representation of tbe constraint
    """
    prefix = NotImplemented


    @property
    def __attrname__(self):
        """
        Name the attribute the instances will have
        Example: GenericConstraint -> generic_constraint
        :return:
        """
        return camel2underscores(self.__class__.__name__)

    def __init__(self, expr, id_='', model=None, hook = None, queue=False, **kwargs):
        """

        :param id_: will be used to identify the variable
            (name will be a concat of this and a prefix)
        :param model: the cobra.Model object
        :param queue: whether or not to queue the variable for update object
        :param kwargs: stuff you want to pass to the variable constructor
        """
        self.hook = hook
        self._id = id_
        self._model = model
        self.kwargs = kwargs
        self._name = self.make_name()
        self.get_interface(expr, queue)
        self.prefix = ''

    def get_interface(self, expr, queue):
        """
        Called upon completion of __init__, initializes the value of self.var,
        which is returned upon call, and stores the actual interfaced variable.

        :return: instance of Variable from the problem
        """
        if not self.name in self.model.constraints:
            constraint = self.model.problem.Constraint(expression = expr,
                                                       name = self.name,
                                                       **self.kwargs)
            if not queue:
                self.model.add_cons_vars(constraint, sloppy=self.model.sloppy)
            else:
                self.model._cons_queue.append(constraint)
        else:
            self.constraint = self.model.constraints.get(self.name)


    def make_name(self):
        """
        Needs to be overridden by the subclass, concats the id with a
         prefix

        :return: None
        """
        return self.prefix + self.id

    def change_expr(self, new_expr, sloppy=False):

        lb = self.constraint.lb
        ub = self.constraint.ub
        name = self.name

        # Remove former constraint to override it
        self.model.solver.remove(name)
        new_cons = self.model.solver.interface.Constraint(name = name,
                                                   expression = new_expr,
                                                   ub = ub,
                                                   lb = lb)
        # Add the new variant
        self.model.solver.add(new_cons, sloppy=sloppy)

    @property
    def expr(self):
        return self.constraint.expression

    @expr.setter
    def expr(self,value):
        self.constraint.expression = value

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        self._name = value

    @property
    def id(self):
        """
        for cobra.thermo.DictList compatibility
        :return:
        """
        return self._id

    @property
    def constraint(self):
        return self.model.constraints[self.name]

    @constraint.setter
    def constraint(self, value):
        self.model.constraints[self.name] = value

    @property
    def model(self):
        return self._model

    def __repr__(self):
        return self.name + ': ' + self.constraint.expression.__repr__()

class ModelConstraint(GenericConstraint):
    """
    Class to represent a variable attached to the model
    """

    def __init__(self, model, expr, id_, **kwargs):
        GenericConstraint.__init__(self,
                                   id_= id_,
                                   expr=expr,
                                   model=model,
                                   hook=model,
                                   **kwargs)

    prefix = 'MODC_'


class GeneConstraint(GenericConstraint):
    """
    Class to represent a variable attached to a enzyme
    """

    def __init__(self, gene, expr, **kwargs):
        model = gene.model

        GenericConstraint.__init__(self,
                                   expr=expr,
                                   model=model,
                                   hook=gene,
                                   **kwargs)

    @property
    def gene(self):
        return self.hook

    @property
    def id(self):
        return self.gene.id

    @property
    def model(self):
        return self.gene.model

    prefix = 'GC_'

class ReactionConstraint(GenericConstraint):
    """
    Class to represent a variable attached to a reaction
    """

    def __init__(self, reaction, expr, **kwargs):
        model = reaction.model

        GenericConstraint.__init__(self,
                                   expr=expr,
                                   model=model,
                                   hook=reaction,
                                   **kwargs)


    @property
    def reaction(self):
        return self.hook

    @property
    def id(self):
        return self.reaction.id

    @property
    def model(self):
        return self.reaction.model

    prefix = 'RC_'

class MetaboliteConstraint(GenericConstraint):
    """
    Class to represent a variable attached to a enzyme
    """

    def __init__(self, metabolite, expr, **kwargs):
        model = metabolite.model

        GenericConstraint.__init__(self,
                                   expr=expr,
                                   model=model,
                                   hook=metabolite
                                   **kwargs)

    @property
    def metabolite(self):
        return self.hook

    @property
    def id(self):
        return self.metabolite.id

    @property
    def model(self):
        return self.metabolite.model

    prefix = 'MC_'

class NegativeDeltaG(ReactionConstraint):
    """
    Class to represent thermodynamics constraints.

    G: - DGR_rxn + DGoRerr_Rxn + RT * StoichCoefProd1 * LC_prod1
       + RT * StoichCoefProd2 * LC_prod2
       + RT * StoichCoefSub1 * LC_subs1
       + RT * StoichCoefSub2 * LC_subs2
       - ...
     = 0
    """

    prefix = 'G_'

class ForwardDeltaGCoupling(ReactionConstraint):
    """
    Class to represent thermodynamics coupling: DeltaG of reactions has to be
    DGR < 0 for the reaction to proceed forwards
    Looks like:
    FU_rxn: 1000 FU_rxn + DGR_rxn < 1000
    """

    def __init__(self, reaction, expr, **kwargs):
        ReactionConstraint.__init__(self, reaction, expr, **kwargs)

    prefix = 'FU_'

class BackwardDeltaGCoupling(ReactionConstraint):
    """
    Class to represent thermodynamics coupling: DeltaG of reactions has to be
    DGR > 0 for the reaction to proceed backwards
    Looks like:
    BU_rxn: 1000 BU_rxn - DGR_rxn < 1000
    """

    def __init__(self, reaction, expr, **kwargs):
        ReactionConstraint.__init__(self, reaction, expr, **kwargs)

    prefix = 'BU_'

class ForwardDirectionCoupling(ReactionConstraint):
    """
    Class to represent a forward directionality coupling with thermodynamics on
    reaction variables
    Looks like :
    UF_rxn: F_rxn - M FU_rxn < 0
    """

    def __init__(self, reaction, expr, **kwargs):
        ReactionConstraint.__init__(self, reaction, expr, **kwargs)

    prefix = 'UF_'


class BackwardDirectionCoupling(ReactionConstraint):
    """
    Class to represent a backward directionality coupling with thermodynamics on
    reaction variables
    Looks like :
    UR_rxn: R_rxn - M RU_rxn < 0
    """

    def __init__(self, reaction, expr, **kwargs):
        ReactionConstraint.__init__(self, reaction, expr, **kwargs)

    prefix = 'UR_'

class SimultaneousUse(ReactionConstraint):
    """
    Class to represent a simultaneous use constraint on reaction variables
    Looks like:
    SU_rxn: FU_rxn + BU_rxn <= 1
    """

    prefix = 'SU_'

class DisplacementCoupling(ReactionConstraint):
    """
    Class to represent the coupling to the thermodynamic displacement
    Looks like:
    Ln(Gamma) - (1/RT)*DGR_rxn = 0
    """

    prefix = 'DC_'

class ForbiddenProfile(GenericConstraint):
    """
    Class to represent a forbidden net flux directionality profile
    Looks like:
    FU_rxn_1 + BU_rxn_2 + ... + FU_rxn_n <= n-1
    """

    def __init__(self, model, expr, id_, **kwargs):

        GenericConstraint.__init__(self,
                                   id_=id_,
                                   expr=expr,
                                   model=model,
                                   **kwargs)

    prefix = 'FP_'


class LinearizationConstraint(ModelConstraint):
    """
    Class to represent a variable attached to a reaction
    """
    @staticmethod
    def from_constraints(cons, model):
        return LinearizationConstraint(
            name = cons.name,
            expr = cons.expr,
            model = model,
            ub = cons.ub,
            lb = cons.lb,
        )

    prefix = 'LC_'


# -*- coding: utf-8 -*-
"""
.. module:: pytfa
   :platform: Unix, Windows
   :synopsis: Thermodynamics-based Flux Analysis

.. moduleauthor:: pyTFA team

Model class
"""

from abc import ABC, abstractmethod
from collections import defaultdict

import pandas as pd
from numpy import empty
import optlang
from optlang.exceptions import SolverError
from cobra import DictList, Model
from cobra.core.solution import Solution

from ..utils.str import camel2underscores
from ..optim.variables import GenericVariable, ReactionVariable, MetaboliteVariable
from ..optim.constraints import ReactionConstraint, MetaboliteConstraint
from ..optim.utils import get_primal, get_all_subclasses

import time

def timeit(method):
    """
    Adapted from Andreas Jung's `blog
    <https://www.zopyx.com/andreas-jung/contents/a-python-decorator-for-measuring-the-execution-time-of-methods>`_

    :param method: The method to time
    :return:
    """


    def timed(self, *args, **kw):
        ts = time.time()
        result = method(self, *args, **kw)
        te = time.time()

        message = '%r (%r, %r) %2.2f sec' % (method.__name__, args, kw, te-ts)

        try:
            self.logger.debug(message)
        except AttributeError:
            print(message)
        return result

    return timed

class LCSBModel(ABC):

    # @abstractmethod
    def __init__(self, model, name, sloppy=False):

        """
        Very much model specific
        """

        Model.__init__(self, model.copy(), name)

        self._cons_queue = list()
        self._var_queue = list()

        self._var_dict = dict()
        self._cons_dict = dict()

        self.sloppy=sloppy


    @abstractmethod
    def copy(self):
        """
        Needs to be reimplemented, as our objects have complicated hierarchy
        :return:
        """

    def print_info(self):
        """
        Print information and counts for the cobra_model
        :return:
        """

        n_metabolites = len(self.metabolites)
        n_reactions = len(self.reactions)
        n_constraints = len(self.constraints)
        n_variables = len(self.variables)

        info = pd.DataFrame(columns=['value'])
        info.loc['name'] = self.name
        info.loc['description'] = self.description
        info.loc['num constraints'] = n_constraints
        info.loc['num variables'] = n_variables
        info.loc['num metabolites'] = n_metabolites
        info.loc['num reactions'] = n_reactions
        info.index.name = 'key'

        print(info)

    def add_variable(self, kind, hook, queue=False, **kwargs):
        """ Add a new variable to a COBRApy cobra_model.

        :param kind:
        :param string,cobra.Reaction hook: Either a string representing the name
            of the variable to add to the cobra_model, or a reaction object if the
            kind allows it

        :returns: The created variable
        :rtype: optlang.interface.Variable

        """

        # Initialisation links to the cobra_model
        var = kind(hook,
                   # lb=lower_bound if lower_bound != float('-inf') else None,
                   # ub=upper_bound if upper_bound != float('inf') else None,
                   queue=queue,
                   **kwargs)

        self._var_dict[var.name] = var
        self.logger.debug('Added variable: {}'.format(var.name))
        # self.add_cons_vars(var.variable)

        return var

    def add_constraint(self, kind, hook, expr, queue=False,**kwargs):
        """ Add a new constraint to a COBRApy cobra_model

        :param kind:
        :param string,cobra.Reaction hook: Either a string representing the name
            of the variable to add to the cobra_model, or a reaction object if the
            kind allows it
        :param sympy.thermo.expr.Expr expr: The expression of the constraint

        :returns: The created constraint
        :rtype: optlang.interface.Constraint

        """

        if isinstance(expr, GenericVariable):
            # make sure we actually pass the optlang variable
            expr = expr.variable

        # Initialisation links to the cobra_model
        cons = kind(hook, expr, # problem = self.problem,
                    # lb=lower_bound if lower_bound != float('-inf') else None,
                    # ub=upper_bound if upper_bound != float('inf') else None,
                    queue=queue,
                    **kwargs)
        self._cons_dict[cons.name] = cons
        self.logger.debug('Added constraint: {}'.format(cons.name))
        # self.add_cons_vars(cons.constraint)

        return cons

    def remove_reactions(self, reactions, remove_orphans=False):
        # Remove the constraints and variables associated to these reactions
        all_cons_subclasses = get_all_subclasses(ReactionConstraint)
        all_var_subclasses = get_all_subclasses(ReactionVariable)

        self._remove_associated_consvar(all_cons_subclasses, all_var_subclasses,
                                        reactions)

        Model.remove_reactions(self,reactions,remove_orphans)

    def remove_metabolites(self, metabolite_list, destructive=False):
        # Remove the constraints and variables associated to these reactions
        all_cons_subclasses = get_all_subclasses(MetaboliteConstraint)
        all_var_subclasses = get_all_subclasses(MetaboliteVariable)

        self._remove_associated_consvar(all_cons_subclasses, all_var_subclasses,
                                        metabolite_list)

        Model.remove_metabolites(self, metabolite_list, destructive)

    def _remove_associated_consvar(self, all_cons_subclasses, all_var_subclasses,
                                   collection):
        """
        Removes both the constraints and variables associated to an element,
        as long as it was used as a hook in the cons/var declaration.
        For example, upon removing a reaction, also removes its associated
        deltaG variables and coupling constraints
        """

        if not hasattr(collection, '__iter__'):
            collection = [collection]

        strfy = lambda x:x if isinstance(x, str) else x.id

        for cons_type in all_cons_subclasses:
            for element in collection:
                try:
                    cons = self._cons_kinds[cons_type.__name__].get_by_id(strfy(element))
                    self.remove_constraint(cons)
                except KeyError as e:
                    pass
        for var_type in all_var_subclasses:
            for element in collection:
                try:
                    var = self._var_kinds[var_type.__name__].get_by_id(strfy(element))
                    self.remove_variable(var)
                except KeyError as e:
                    pass


    def remove_variable(self, var):
        """
        Removes a variable

        :param var:
        :return:
        """
        # Get the pytfa var object if an optlang variable is passed
        if isinstance(var,optlang.Variable):
            var = self._var_dict[var.name]

        self._var_dict.pop(var.name)
        self.remove_cons_vars(var.variable)
        self.logger.debug('Removed variable {}'.format(var.name))

    def remove_constraint(self, cons):
        """
        Removes a constraint

        :param cons:
        :return:
        """
        # Get the pytfa var object if an optlang variable is passed
        if isinstance(cons,optlang.Constraint):
            cons = self._cons_dict[cons.name]

        self._cons_dict.pop(cons.name)
        self.remove_cons_vars(cons.constraint)
        self.logger.debug('Removed constraint {}'.format(cons.name))

    def _push_queue(self):
        """
        updates the constraints and variables of the model with what's in the
        queue
        :return:
        """

        self.add_cons_vars(self._var_queue, sloppy=self.sloppy)
        self.add_cons_vars(self._cons_queue, sloppy = self.sloppy)

        if len(self._var_queue) > 0:
            self.regenerate_variables()
        if len(self._cons_queue) > 0:
            self.regenerate_constraints()

        self._var_queue = list()
        self._cons_queue = list()


    def regenerate_variables(self):
        """
        Generates references to the cobra_model's constraints in self._var_dict
        as tab-searchable attributes of the thermo cobra_model
        :return:
        """

        # Let us not forget to remove fields that might be empty by now
        if hasattr(self, '_var_kinds'):
            for k in self._var_kinds:
                attrname = camel2underscores(k)
                try:
                    delattr(self, attrname)
                except AttributeError:
                    pass # The attribute may not have been set up yet

        _var_kinds = defaultdict(DictList)
        for k, v in self._var_dict.items():
            _var_kinds[v.__class__.__name__].append(v)

        for k in _var_kinds:
            attrname = camel2underscores(k)
            setattr(self, attrname, _var_kinds[k])

        self._var_kinds = _var_kinds

    def regenerate_constraints(self):
        """
        Generates references to the cobra_model's constraints in self._cons_dict
        as tab-searchable attributes of the thermo cobra_model
        :return:
        """

        # Let us not forget to remove fields that migh be empty by now
        if hasattr(self, '_cons_kinds'):
            for k in self._cons_kinds:
                attrname = camel2underscores(k)
                try:
                    delattr(self, attrname)
                except AttributeError:
                    pass # The attribute may not have been set up yet

        _cons_kinds = defaultdict(DictList)

        for k, v in self._cons_dict.items():
            _cons_kinds[v.__class__.__name__].append(v)

        for k in _cons_kinds:
            attrname = camel2underscores(k)
            setattr(self, attrname, _cons_kinds[k])

        self._cons_kinds = _cons_kinds

    def repair(self):
        """
        Updates references to variables and constraints
        :return:
        """
        # self.add_cons_vars([x.constraint for x in self._cons_dict.values()])
        # self.add_cons_vars([x.variable for x in self._var_dict.values()])
        self._push_queue()
        Model.repair(self)
        self.regenerate_constraints()
        self.regenerate_variables()

    def get_primal(self, vartype, index_by_reactions=False):
        """
        Returns the primal value of the cobra_model for variables of a given type

        :param index_by_reactions:
        :param vartype: Class of variable. Ex: pytfa.optim.variables.ThermoDisplacement
        :return:
        """
        return get_primal(self, vartype, index_by_reactions)

    def get_solution(self):
        """
        Overrides the cobra.thermo.solution method, to also get the supplementary
        variables we added to the cobra_model

        *   :code:`solution.fluxes` in `cobrapy` is a transformed version of the solver
            output, as it actually calculates the _net_ flux of each reaction by
            substracting the reverse variable value to the forward variable value.
            This should be used anytime one needs the actual flux value

        *   :code:`solution.raw` is a clear copy of the solver output. From there one
            can access the value at solution for all the variables of the problem.
            However, looking for a reaction ID in there will only give the
            _forward_ flux. This should be used for any other variable than fluxes.

        *   :code:`solution.values` yields variables multiplied by their scaling factor
            (1 by default). Useful if you operated scaling on your equations for
            numerical reasons. This does _not_ include fluxes

        :return:
        """
        objective_value = self.solver.objective.value
        status = self.solver.status
        variables = pd.Series(data=self.solver.primal_values)

        fluxes = empty(len(self.reactions))
        rxn_index = list()
        var_primals = self.solver.primal_values

        for (i, rxn) in enumerate(self.reactions):
            rxn_index.append(rxn.id)
            fluxes[i] = var_primals[rxn.id] - var_primals[rxn.reverse_id]

        fluxes = pd.Series(index=rxn_index, data=fluxes, name="fluxes")

        solution = Solution(objective_value=objective_value, status=status,
                            fluxes=fluxes)

        self.solution = solution

        self.solution.raw = variables

        self.\
            solution.values = pd.DataFrame.from_dict({k:v.unscaled
                                                 for k,v in self._var_dict.items()},
                                                 orient = 'index')

        return solution

    def optimize(self, objective_sense=None, **kwargs):
        """
        Call the Model.optimize function (which really is but an interface to the
        solver's. Catches SolverError in the case of no solutions. Passes down
        supplementary keyword arguments (see cobra.thermo.Model.optimize)
        :type objective_sense: 'min' or 'max'
        """

        if objective_sense:
            self.objective.direction = objective_sense

        try:
            # self._hidden_optimize_call(kwargs)
            Model.optimize(self, **kwargs)
            solution = self.get_solution()
            self.solution = solution
            return solution
        except SolverError as SE:
            status = self.solver.status
            self.logger.error(SE)
            self.logger.warning('Solver status: {}'.format(status))
            raise (SE)

    # @timeit
    # def _hidden_optimize_call(self, kwargs):
    #     return Model.optimize(self, **kwargs)

    @timeit
    def slim_optimize(self, *args, **kwargs):
        return Model.slim_optimize(self, *args, **kwargs)

    def get_constraints_of_type(self, constraint_type):
        """
        Convenience function that takes as input a constraint class and returns
        all its instances within the cobra_model

        :param constraint_type:
        :return:
        """
        if isinstance(constraint_type,str):
            constraint_key = constraint_type
        else:
            #it is a class
            constraint_key = constraint_type.__name__
        return self._cons_kinds[constraint_key]

    def get_variables_of_type(self, variable_type):
        """
        Convenience function that takes as input a variable class and returns
        all its instances within the cobra_model

        :param variable_type:
        :return:
        """
        if isinstance(variable_type,str):
            variable_key = variable_type
        else:
            #it is a class
            variable_key = variable_type.__name__
        return self._var_kinds[variable_key]



# -*- coding: utf-8 -*-
"""Thermodynamic information for metabolites from eQuilibrator.

.. module:: pytfa
   :platform: Unix, Windows
   :synopsis: Thermodynamics-based Flux Analysis

.. moduleauthor:: pyTFA team
"""

import equilibrator_cache.compatibility as compat

# this file is not imported in the __init__.py. The explicit import of the file
# triggers the loading of eQuilbrator
from equilibrator_api import ComponentContribution, Q_
from equilibrator_api.phased_reaction import PhasedReaction
from equilibrator_cache import create_compound_cache_from_quilt
from .std import TEMPERATURE_0
from ..utils.numerics import BIGM_DG
from ..utils.logger import get_bistream_logger

# load compound cache just one to aliviate resource usage
ccache = None
logger = get_bistream_logger("eQuilibrator_formation")


def build_thermo_from_equilibrator(model, T=TEMPERATURE_0):
    """Build `thermo_data` structure from a cobra Model.

    The structure of the returned dictionary is specified in the pyTFA
    [documentation](https://pytfa.readthedocs.io/en/latest/thermoDB.html).

    :param model: cobra.Model
    :return thermo_data: dict
        to be passed as argument to initialize a `ThermoModel`.

    """
    global ccache
    if ccache is None:
        ccache = create_compound_cache_from_quilt()
        logger.debug("eQuilibrator compound cache is loaded.")

    cc = ComponentContribution()
    cc.temperature=Q_(str(T) + "K")

    thermo_data = {"name": "eQuilibrator", "units": "kJ/mol", "cues": {}}
    met_to_comps = compat.map_cobra_metabolites(ccache, model.metabolites)
    thermo_data["metabolites"] = [
        compound_to_entry(met, cc) for met in met_to_comps
    ]
    return thermo_data


def compute_dGf(compound, cc):
    """Get ΔGf from equilibrator `compound`."""
    dG0_prime, dG0_uncertainty = cc.dG0_prime(
        PhasedReaction(sparse={compound: 1}, rid="tmp_"+compound.id)
    )
    return dG0_prime, dG0_uncertainty


def compound_to_entry(compound, cc):
    """Build thermo structure entry from a `equilibrator_cache.Compound`.

    eQuilibrator works with Component Contribution instead of groups, so it is
    not possible to generate cues from it.

    :param compound: equilibrator_cache.Compound
    :return: dict
        with keys ['deltaGf_std', 'deltaGf_err', 'error', 'struct_cues',
        'id', 'pKa', 'mass_std', 'charge_std', 'nH_std', 'name', 'formula',
        'other_names']

    """
    deltaGf_std, deltaGf_err = (BIGM_DG, BIGM_DG)
    nH_std = compound.atom_bag["H"] if "H" in compound.atom_bag else 0
    try:
        deltaGf_std, deltaGf_err = compute_dGf(compound, cc)
        err = "Nil"
    except Exception as e:
        err = 1
        logger.debug(
            "{} : thermo data NOT created, error : {}".format(
                compound.id, e
            )
        )
    return dict(
        deltaGf_std=deltaGf_std,
        deltaGf_err=deltaGf_err,
        error=err,
        struct_cues=None,
        id=compound.id,
        pKa=compound.dissociation_constants,
        mass_std=compound.mass,
        charge_std=None,
        nH_std=nH_std,
        name=compound.id,
        formula=compound.formula,
        other_names=compound.identifiers,
    )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
.. module:: redgem
   :platform: Unix, Windows
   :synopsis: RedGEM Algorithm

.. moduleauthor:: pyTFA team

Model class
"""

import networkx as nx
from cobra import Metabolite, Reaction, Model
from copy import deepcopy


class NetworkExpansion:

    def __init__(self, gem, core_subsystems, extracellular_system,
                 cofactors, small_metabolites, inorganics,
                 d, n):
        """
        A class encapsulating the RedGEM algorithm

        :param gem: The studied GEM
        :param core_subsystems: Core subsystems
        :param extracellular_system: Extracellular metabolite ids
        :param cofactors: List of cofactors id
        :param small_metabolites: List of small metabolites id
        :param inorganics: List of inorganics id
        :param d: Degree
        :param n: User parameter
        """
        # Shallow copy of the GEM : the deepcopy is possibly performed in redgem, before
        # calling NetworkExpansion
        self._redgem = gem
        #self._redgem.name = 'redgem'
        self._graph = nx.DiGraph()

        # Subsystems
        self._core_subsystems = core_subsystems
        self._subsystem_count = len(core_subsystems)
        self._extracellular_system = extracellular_system

        # Dicts to save extracted reactions and metabolites for each subsystem
        # TODO: Improve structure definition
        dict_of_lists_of_sets = {}
        for name in core_subsystems:
            dict_of_lists_of_sets[name] = [set() for _ in range(d+1)]
        dict_of_dicts_of_lists_of_sets = {}
        for name in core_subsystems:
            dict_of_dicts_of_lists_of_sets[name] = deepcopy(dict_of_lists_of_sets)
        dict_of_int = {}
        for name in core_subsystems:
            dict_of_int[name] = -1
        dict_of_dicts_of_int = {}
        for name in core_subsystems:
            dict_of_dicts_of_int[name] = deepcopy(dict_of_int)

        self._subsystem_reactions = {}
        self._subsystem_reactions_id = {}
        self._intermediate_reactions_id = deepcopy(dict_of_dicts_of_lists_of_sets)
        self._subsystem_metabolites = {}
        self._subsystem_metabolites_id = {}
        self._intermediate_metabolites_id = deepcopy(dict_of_dicts_of_lists_of_sets)
        self._intermediate_paths = deepcopy(dict_of_dicts_of_lists_of_sets)
        self._min_distance_sub_to_sub = deepcopy(dict_of_dicts_of_int)

        self._intermediate_extracellular_paths = deepcopy(dict_of_lists_of_sets)
        self._intermediate_extracellular_metabolites_id = deepcopy(dict_of_lists_of_sets)
        self._intermediate_extracellular_reactions_id = deepcopy(dict_of_lists_of_sets)

        self._path_dict = {}

        # Save others parameters
        self._cofactor_pairs = cofactors
        self._small_metabolites = small_metabolites
        self._inorganics = inorganics
        self._d = d
        self._n = n

    def extract_subsystem_reactions(self, subsystem):
        """
        Extracts all reactions of a subsystem and stores them and their id in the corresponding
        dictionary.

        :param subsystem: Name of the subsystem
        :return: Extracted reactions
        """
        rxns = set()
        rxns_id = set()
        for rxn in self._redgem.reactions:
            if rxn.subsystem == subsystem:
                rxns.add(rxn)
                rxns_id.add(rxn.id)
        self._subsystem_reactions[subsystem] = rxns
        self._subsystem_reactions_id[subsystem] = rxns_id
        return rxns

    def extract_subsystem_metabolites(self, subsystem):
        """
        Extracts all metabolites of a subsystem and stores them and their id in the corresponding
        dictionary.

        :param subsystem: Name of the subsystem
        :return: Extracted metabolites
        """
        subsystem_rxns = self._subsystem_reactions[subsystem]
        metabolites = set()
        metabolites_id = set()
        for rxn in subsystem_rxns:
            for metabolite in rxn.metabolites:
                metabolite_id = metabolite.id
                if metabolite_id in self._cofactor_pairs \
                        or metabolite_id in self._small_metabolites \
                        or metabolite_id in self._inorganics:
                    continue
                metabolites.add(metabolite)
                metabolites_id.add(metabolite.id)
        self._subsystem_metabolites[subsystem] = metabolites
        self._subsystem_metabolites_id[subsystem] = metabolites_id
        return metabolites

    def create_new_stoichiometric_matrix(self):
        """
        Extracts the new graph without the small metabolites, inorganics and cofactor pairs.

        :return: Networkx graph of the new network
        """
        kept_rxns = []
        kept_metabolites = set()
        for rxn in self._redgem.reactions:
            metabolites = {}
            for metabolite, coefficient in rxn.metabolites.items():
                metabolite_id = metabolite.id
                if metabolite_id in self._cofactor_pairs \
                        or metabolite_id in self._small_metabolites \
                        or metabolite_id in self._inorganics:
                    continue
                new_metabolite = Metabolite(metabolite_id,
                                            formula=metabolite.formula,
                                            name=metabolite.name,
                                            compartment=metabolite.compartment)
                metabolites[new_metabolite] = coefficient
                kept_metabolites.add(metabolite)
            new_rxn = Reaction(rxn.id,
                               name=rxn.name,
                               subsystem=rxn.subsystem,
                               lower_bound=rxn.lower_bound,
                               upper_bound=rxn.upper_bound)
            new_rxn.add_metabolites(metabolites)
            kept_rxns.append(new_rxn)

        paths_struct = [{} for _ in range(self._d+1)]  # Comprehension list to create multiple dicts
        to_struct = [""] * (self._d+1)
        for metabolite in kept_metabolites:
            self._graph.add_node(metabolite.id, paths=paths_struct, to=to_struct)
        for rxn in kept_rxns:
            for reactant in rxn.reactants:
                for product in rxn.products:
                    self._graph.add_edge(reactant.id, product.id, rxn_id=rxn.id, weight=1)
        return self._graph

    def breadth_search_subsystems_paths_length_d(self, subsystem_i, subsystem_j, d):
        """
        Breadth first search from each metabolite in subsystem i with special stop conditions
        during exploration for paths of length d.

        This function explores the graph through allowed paths only : this path can't go through
        subsystem i or j but must start in i and end in j. The length of each path found is d.

        :param subsystem_i: Source subsystem
        :param subsystem_j: Destination subsystem
        :param d: Path length desired
        :return: None
        """
        for metabolite_id in self._subsystem_metabolites_id[subsystem_i]:
            # Find metabolites at a distance d from metabolite_id
            ancestors = {}
            frontier = {metabolite_id}
            explored = {metabolite_id}
            for i in range(d):
                new_nodes = set()
                for current_node in frontier:
                    for new_node in set(self._graph.adj[current_node]):
                        if self.is_node_allowed(new_node, i, explored, subsystem_i, subsystem_j, d):
                            new_nodes.add(new_node)
                            # new_node can already be in ancestors if there are 2 paths of same
                            # length to it
                            if new_node in ancestors:
                                ancestors[new_node].append(current_node)
                            else:
                                ancestors[new_node] = [current_node]
                explored = explored.union(new_nodes)
                frontier = new_nodes
            # Handle d = 0 case, since it didn't go through the loop
            if d == 0 and metabolite_id not in self._subsystem_metabolites_id[subsystem_j]:
                frontier = {}
            # Retrieve and save metabolites, reactions and paths
            for node in frontier:
                paths = self.retrieve_all_paths(node, metabolite_id, ancestors)
                self._intermediate_paths[subsystem_i][subsystem_j][d] = \
                    self._intermediate_paths[subsystem_i][subsystem_j][d].union(set(paths))
                self.retrieve_intermediate_metabolites_and_reactions(paths, subsystem_i,
                                                                     subsystem_j, d)

    def is_node_allowed(self, node, i, explored, subsystem_i, subsystem_j, d):
        """
        Checks whether or not a metabolite is allowed for the current path.

        The new node is added if it is not already explored, if it is not in the source subsystem,
        and if it is not in the destination subsystem, except if it is the last round
        of exploration

        :param node: Metabolite id
        :param i: Current step
        :param explored: Explored node for this path
        :param subsystem_i: Source subsystem
        :param subsystem_j: Destination subsystem
        :param d: Path length desired
        :return: Boolean answering the question
        """
        if node in explored:
            return False
        if subsystem_i != subsystem_j and node in self._subsystem_metabolites_id[subsystem_i]:
            return False
        if i < d-1 and node in self._subsystem_metabolites_id[subsystem_j]:
            return False
        if i == d-1 and node not in self._subsystem_metabolites_id[subsystem_j]:
            return False
        return True

    def retrieve_all_paths(self, dest_node, src_node, ancestors, init_dict=True):
        """
        Retrieves all paths between a source metabolite and a destination metabolite after a
        breadth first search.

        This function is a recursive function, which makes use of dynamic programming to reduce
        its complexity. It uses self._path_dict to store already computed data.

        :param dest_node: Destination metabolite
        :param src_node: Source metabolite
        :param ancestors: Dictionary with ancestors found during the search
        :param init_dict: Boolean, for function initialisation
        :return: A list of all paths as tuples
        """
        if init_dict:
            self._path_dict = {}
        if dest_node == src_node:
            self._path_dict[dest_node] = [(src_node,)]
        if dest_node not in self._path_dict:
            new_paths = []
            for previous_node in ancestors[dest_node]:
                for path in self.retrieve_all_paths(previous_node, src_node, ancestors, False):
                    new_paths.append(path + (dest_node,))
            self._path_dict[dest_node] = new_paths
        return self._path_dict[dest_node]

    def retrieve_intermediate_metabolites_and_reactions(self, paths, subsystem_i, subsystem_j, d):
        """
        Retrieves and stores intermediate metabolites and reactions (i.e. M_{i,j}, R_{i,j},
        M_{i,i} and R_{i,i}).

        This function adds all reactions contained in these paths, and all metabolites between

        :param paths: List of paths between subsystems
        :param subsystem_i: Source subsystem
        :param subsystem_j: Destination subsystem
        :param d: Path length
        :return: None
        """
        for path in paths:
            for i in range(len(path)-1):
                reaction = self._graph[path[i]][path[i+1]]['rxn_id']
                self._intermediate_reactions_id[subsystem_i][subsystem_j][d].add(reaction)
                if i > 0:
                    self._intermediate_metabolites_id[subsystem_i][subsystem_j][d].add(path[i])

    def find_min_distance_between_subsystems(self):
        """
        Find minimal distance between each subsystems in both directions

        :return: Dict with distances
        """
        for i in self._core_subsystems:
            for j in self._core_subsystems:
                for k in range(self._d+1):
                    # If there path of length d
                    if self._intermediate_paths[i][j][k]:
                        self._min_distance_sub_to_sub[i][j] = k
                        break
                # If min distance os not found, then
                if self._min_distance_sub_to_sub[i][j] == -1:
                    pass
        return self._min_distance_sub_to_sub

    def breadth_search_extracellular_system_paths(self, subsystem, n):
        """
        Breadth first search from each metabolite in the extracellular system with special stop
        conditions during exploration for paths of length n.

        This function explores the graph through allowed paths only : this path can't go through
        the extracellular system or the subsystem but must start in the extracellular system and
        end in the subsystem. The length of each path found is n.

        :param subsystem: Destination subsystem
        :param n: Path length desired
        :return: None
        """
        for metabolite_id in self._extracellular_system:
            # Find metabolites at a distance n from metabolite_id
            if metabolite_id not in self._graph:
                continue
            ancestors = {}
            frontier = {metabolite_id}
            explored = {metabolite_id}
            for i in range(n):
                new_nodes = set()
                for current_node in frontier:
                    for new_node in set(self._graph.adj[current_node]):
                        if self.is_node_allowed_extracellular(new_node, i, explored, subsystem, n):
                            new_nodes.add(new_node)
                            # new_node can already be in ancestors if there are 2 paths of same
                            # length to it
                            if new_node in ancestors:
                                ancestors[new_node].append(current_node)
                            else:
                                ancestors[new_node] = [current_node]
                explored = explored.union(new_nodes)
                frontier = new_nodes
            # Handle n = 0 case, since it didn't go through the loop
            if n == 0 and metabolite_id not in self._subsystem_metabolites_id[subsystem]:
                frontier = {}
            # Retrieve and save metabolites, reactions and paths
            for node in frontier:
                paths = self.retrieve_all_paths(node, metabolite_id, ancestors)
                self._intermediate_extracellular_paths[subsystem][n] = \
                    self._intermediate_extracellular_paths[subsystem][n].union(set(paths))
                self.retrieve_intermediate_extracellular_metabolites_and_reactions(paths, subsystem,
                                                                                   n)

    def is_node_allowed_extracellular(self, node, i, explored, subsystem, n):
        """
        Checks whether or not a metabolite is allowed for the current path.

        The new node is added if it is not already explored, if it is not in the extracellular
        system, and if it is not in the destination subsystem except if it is the last round
        of exploration

        :param node: Metabolite id
        :param i: Current step
        :param explored: Explored node for this path
        :param subsystem: Destination subsystem
        :param n: Path length desired
        :return: Boolean answering the question
        """
        if node in explored:
            return False
        if node in self._extracellular_system:
            return False
        if i < n-1 and node in self._subsystem_metabolites_id[subsystem]:
            return False
        if i == n-1 and node not in self._subsystem_metabolites_id[subsystem]:
            return False
        return True

    def retrieve_intermediate_extracellular_metabolites_and_reactions(self, paths, subsystem, n):
        """
        Retrieves and stores intermediate metabolites and reactions for the extracellular system

        This function adds all reactions contained in these paths, and all metabolites between

        :param paths: List of paths
        :param subsystem: Destination subsystem
        :param n: Path length
        :return: None
        """
        for path in paths:
            for i in range(len(path) - 1):
                reaction = self._graph[path[i]][path[i + 1]]['rxn_id']
                self._intermediate_extracellular_reactions_id[subsystem][n].add(reaction)
                if i > 0:
                    self._intermediate_extracellular_metabolites_id[subsystem][n].add(path[i])

    def run_between_all_subsystems(self):
        """
        Retrieve subsystem and intermediate reactions and metabolites.

        :return: None
        """
        for subsystem in self._core_subsystems:
            self.extract_subsystem_reactions(subsystem)
            self.extract_subsystem_metabolites(subsystem)

        for subsystem_i in self._core_subsystems:
            for subsystem_j in self._core_subsystems:
                for k in range(self._d+1):
                    self.breadth_search_subsystems_paths_length_d(subsystem_i, subsystem_j, k)

    def run_extracellular_system(self):
        """
        Retrieve intermediate reactions and metabolites for the extracellular system

        :return: None
        """
        for subsystem in self._core_subsystems:
            for k in range(self._n + 1):
                self.breadth_search_extracellular_system_paths(subsystem, k)

    def extract_sub_network(self):
        """
        Extracts the reduced gem.

        :return: None
        """
        def extract_id(x):
            return x.id
        to_remove_metabolites = set(map(extract_id, self._redgem.metabolites))
        to_remove_reactions = set(map(extract_id, self._redgem.reactions))

        # Keep subsystems reactions and metabolites
        for name in self._core_subsystems:
            to_remove_reactions = to_remove_reactions - self._subsystem_reactions_id[name]
            to_remove_metabolites = to_remove_metabolites - self._subsystem_metabolites_id[name]

        # Keep intermediate reactions and metabolites
        for i in self._core_subsystems:
            for j in self._core_subsystems:
                for k in range(self._d+1):
                    to_remove_reactions = to_remove_reactions \
                                          - self._intermediate_reactions_id[i][j][k]
                    to_remove_metabolites = to_remove_metabolites \
                                            - self._intermediate_metabolites_id[i][j][k]

        # Keep extracellular metabolites
        to_remove_metabolites = to_remove_metabolites - set(self._extracellular_system)

        # Keep intermediate extracellular reactions and metabolites
        for i in self._core_subsystems:
            for k in range(self._d+1):
                to_remove_reactions = to_remove_reactions \
                                      - self._intermediate_extracellular_reactions_id[i][k]
                to_remove_metabolites = to_remove_metabolites \
                                        - self._intermediate_extracellular_metabolites_id[i][k]

        self._redgem.remove_reactions(to_remove_reactions, True)

    def run(self):
        """
        Runs RedGEM.

        :return: None
        """
        self.create_new_stoichiometric_matrix()
        self.run_between_all_subsystems()
        self.run_extracellular_system()
        self.extract_sub_network()

        return self._redgem



# -*- coding: utf-8 -*-
"""
.. module:: pytfa
   :platform: Unix, Windows
   :synopsis: Thermodynamics-based Flux Analysis

.. moduleauthor:: pyTFA team

Make the model serializable
"""
from collections import OrderedDict, defaultdict
import cobra.io.dict as cbd
from cobra.exceptions import SolverNotFound

from ..thermo.tmodel import ThermoModel

from ..optim.variables import ReactionVariable, MetaboliteVariable, ModelVariable
from ..optim.constraints import ReactionConstraint, MetaboliteConstraint, ModelConstraint
from ..optim.utils import symbol_sum

from ..thermo.metabolite import MetaboliteThermo

from optlang.util import expr_to_json, parse_expr

from copy import copy

import sympy


def get_all_subclasses(cls):
    all_subclasses = []

    for subclass in cls.__subclasses__():
        all_subclasses.append(subclass)
        all_subclasses.extend(get_all_subclasses(subclass))

    return all_subclasses

def make_subclasses_dict(cls):
    the_dict = {x.__name__:x for x in get_all_subclasses(cls)}
    the_dict[cls.__name__] = cls
    return the_dict

def get_model_variable_subclasses():
   return make_subclasses_dict(ModelVariable)

def get_model_constraint_subclasses():
   return make_subclasses_dict(ModelConstraint)


BASE_NAME2HOOK = {
                    ReactionVariable    :'reactions',
                    ReactionConstraint  :'reactions',
                    MetaboliteVariable  :'metabolites',
                    MetaboliteConstraint:'metabolites',
                }


SOLVER_DICT = {
    'optlang.gurobi_interface':'optlang-gurobi',
    'optlang.cplex_interface':'optlang-cplex',
    'optlang.glpk_interface':'optlang-glpk',
}

def metabolite_thermo_to_dict(metthermo):
    return metthermo.thermo.__dict__

def var_to_dict(variable):
    obj = OrderedDict()
    obj['id'] = variable.id
    obj['name'] = variable.name
    obj['kind'] = type(variable).__name__
    obj['lb'] = variable.variable.lb
    obj['ub'] = variable.variable.ub
    obj['type'] = variable.type

    # For backward compatibility
    try:
        obj['scaling_factor'] = variable.scaling_factor
    except AttributeError:
        obj['scaling_factor'] = 1

    return obj

def cons_to_dict(constraint):
    obj = OrderedDict()
    obj['id'] = constraint.id
    obj['name'] = constraint.name
    obj['kind'] = type(constraint).__name__
    obj['lb'] = constraint.constraint.lb
    obj['ub'] = constraint.constraint.ub
    # obj['expression'] = str(constraint.expr)
    obj['expression'] = expr_to_json(constraint.expr)
    return obj

def archive_variables(var_dict):
    obj = OrderedDict()

    obj['variables'] = []
    for classname,variables in var_dict.items():
        obj[classname] = list(map(var_to_dict, variables))

    return obj

def archive_constraints(cons_dict):
    obj = OrderedDict()

    for classname,constraints in cons_dict.items():
        obj[classname] = list(map(cons_to_dict, constraints))

    return obj

def archive_compositions(compositions):
    """
    Turns a peptide compositions dict of the form:
    { 'b3991': defaultdict(int,
             {<Metabolite ala__L_c at 0x7f7d25504f28>: -42,
              <Metabolite arg__L_c at 0x7f7d2550bcf8>: -11,
              <Metabolite asn__L_c at 0x7f7d2550beb8>: -6,
              ...}),
    ...}


    to:

    { 'b3991': defaultdict(int,,
            {'ala__L_c': -42,
             'arg__L_c': -11,
             'asn__L_c': -6,
              ...}),
    ...}
    :param compositions:
    :return:
    """
    obj = {}
    for k, stoich in compositions.items():
        obj[k] = _stoichiometry_to_dict(stoich)

    return obj

def _stoichiometry_to_dict(stoichiometric_dict):
    """
    Turns a stoichiometric compositions dict of the form:
    'b3991': defaultdict(int,
           {<Metabolite ala__L_c at 0x7f7d25504f28>: -42,
            <Metabolite arg__L_c at 0x7f7d2550bcf8>: -11,
            <Metabolite asn__L_c at 0x7f7d2550beb8>: -6,
            ...})

    to:

    'b3991': defaultdict(int,,
            {'ala__L_c': -42,
             'arg__L_c': -11,
             'asn__L_c': -6,
              ...})
    """
    return defaultdict(int, {k.id:v for k,v in stoichiometric_dict.items()})



def get_solver_string(model):
    return SOLVER_DICT[model.solver.__class__.__module__]


def obj_to_dict(model):
    if model.objective.expression:
        return {x.name: float(v)
                for x,v in model.objective.expression.as_coefficients_dict().items()}
    else:
        return 0

def model_to_dict(model):
    """

    :param model:
    :return:
    """

    # Take advantage of cobra's dict serialization for metabolites and
    # reactions
    obj = cbd.model_to_dict(model)

    obj['solver'] = get_solver_string(model)
    obj['objective'] = obj_to_dict(model)

    # Copy variables, constraints
    # obj['var_dict'] = archive_variables(model._var_kinds)
    # obj['cons_dict'] = archive_constraints(model._cons_kinds)
    obj['variables'] = list(map(var_to_dict, model._var_dict.values()))
    obj['constraints'] = list(map(cons_to_dict, model._cons_dict.values()))

    is_thermo = False

    if isinstance(model, ThermoModel):
        obj['kind'] = 'ThermoModel'
        obj['thermo_data'] = model.thermo_data #it's a dict
        obj['name'] = model.name
        obj['temperature'] = model.TEMPERATURE
        obj['min_ph'] = model.MIN_pH
        obj['max_ph'] = model.MAX_pH
        is_thermo = True

    # Metabolite and Reaction-level cleanup

    for rxn_dict in obj['reactions']:
        rxn = model.reactions.get_by_id(rxn_dict['id'])

        if is_thermo:
            _add_thermo_reaction_info(rxn, rxn_dict)


    # Peptides and Thermo
    for met_dict in obj['metabolites']:
        the_met_id = met_dict['id']
        is_peptide = False

        if is_thermo and not is_peptide: # peptides have no thermo
            the_met = model.metabolites.get_by_id(the_met_id)
            _add_thermo_metabolite_info(the_met, met_dict)
            met_dict['kind'] = 'Metabolite'

    # Relaxation info
    try:
        obj['relaxation'] = model.relaxation
    except AttributeError:
        pass

    return obj


def _add_thermo_reaction_info(rxn, rxn_dict):
    if hasattr(rxn, 'thermo'):
        rxn_dict['thermo'] = rxn.thermo

def _add_thermo_metabolite_info(met, met_dict):
    if hasattr(met, 'thermo'):
        met_dict['thermo'] = metabolite_thermo_to_dict(met)

def model_from_dict(obj, solver=None, custom_hooks = None):
    """
    Custom_hooks looks like

    .. code:: python

        custom_hooks = {<EnzymeVariable Class at 0xffffff> : 'enzymes',
                        ... }

    :param obj:
    :param solver:
    :param custom_hooks:
    :return:
    """
    # Take advantage of cobra's serialization of mets and reactions
    cbm = cbd.model_from_dict(obj)

    if solver is not None:
        try:
            cbm.solver = solver
        except SolverNotFound as snf:
            raise snf
    else:
        try:
            cbm.solver = obj['solver']
        except KeyError:
            pass

    if custom_hooks is None:
        custom_hooks = dict()

    custom_hooks.update(BASE_NAME2HOOK)

    if obj['kind'] == 'ThermoModel':
        new = ThermoModel(thermo_data=obj['thermo_data'],
                          model=cbm,
                          name=obj['name'],
                          temperature=obj['temperature'],
                          min_ph=obj['min_ph'],
                          max_ph=obj['max_ph'])
        new = init_thermo_model_from_dict(new, obj)
    else:
        new = ThermoModel(model=cbm,
                          name=obj['name'])

    new._push_queue()

    name2class, name2hook = add_custom_classes(new,custom_hooks)

    for the_var_dict in obj['variables']:
        this_id = the_var_dict['id']
        classname = the_var_dict['kind']
        lb = the_var_dict['lb']
        ub = the_var_dict['ub']
        scaling_factor = the_var_dict['scaling_factor']

        if classname in get_model_variable_subclasses():
            hook = new
            this_class = get_model_variable_subclasses()[classname]
            nv = new.add_variable(kind=this_class,
                                  hook=hook,
                                  id_=this_id,
                                  ub=ub,
                                  lb=lb,
                                  queue=True,
                                  scaling_factor=scaling_factor)
        elif classname in name2class:
            hook = name2hook[classname].get_by_id(this_id)
            this_class = name2class[classname]
            nv = new.add_variable(kind=this_class,
                                  hook=hook,
                                  ub = ub,
                                  lb = lb,
                                  queue=True,
                                  scaling_factor=scaling_factor)
        else:
            raise TypeError(
                'Class {} serialization not handled yet' \
                    .format(classname))

    new._push_queue()

    variable_parse_dict = {x.name:x for x in new.variables}

    for the_cons_dict in obj['constraints']:
        this_id = the_cons_dict['id']
        classname = the_cons_dict['kind']
        new_expr = parse_expr(the_cons_dict['expression'],
                              local_dict = variable_parse_dict)
        # str_expr = the_cons_dict['expression']
        #
        # # Sympify the expression so that we can substitute variables afterwards
        # sym_expr = sympify(str_expr)
        #
        # subs_dict = {x:new.variables.get(x.name) for x in sym_expr.free_symbols}
        #
        # new_expr = sym_expr.subs(subs_dict)
        lb = the_cons_dict['lb']
        ub = the_cons_dict['ub']

        # Look for the corresponding class:

        if classname in get_model_constraint_subclasses():
            hook=new
            this_class = get_model_constraint_subclasses()[classname]
            nc = new.add_constraint(kind=this_class, hook=hook,
                                    expr=new_expr, id_ = this_id,
                                    ub = ub,
                                    lb = lb,
                                    queue=True)
        elif classname in name2class:
            hook = name2hook[classname].get_by_id(this_id)
            this_class = name2class[classname]
            nc = new.add_constraint(kind=this_class, hook=hook,
                                    expr=new_expr,
                                    ub = ub,
                                    lb = lb,
                                    queue=True)
        else:
            raise TypeError('Class {} serialization not handled yet' \
                            .format(classname))

    new.repair()

    try:
        rebuild_obj_from_dict(new, obj['objective'])
    except KeyError:
        pass

    # Relaxation info
    try:
        new.relaxation = obj['relaxation']
    except KeyError:
        pass

    return new

def rebuild_obj_from_dict(new, objective_dict):
    if objective_dict.__class__ is dict:
        obj_expr = symbol_sum([v*new.variables.get(x) for x,v in objective_dict.items()])
        new.objective = obj_expr
    else:
        new.objective = sympy.S.Zero

def add_custom_classes(model, custom_hooks):
    """
    Allows custom variable serialization

    :param model:
    :param base_classes:
    :param base_hooks:
    :param custom_hooks:
    :return:
    """

    base_classes = dict()
    base_hooks = dict()

    for this_class, hookname in custom_hooks.items():
        # Build the subclass dict of the shape
        # {'MyClass':<MyClass Object>}
        this_subclass_dict = make_subclasses_dict(this_class)
        base_classes.update(this_subclass_dict)

        # Build the Hook dict, of the shape
        # {'MySubClass':model.my_attr}
        hooks = getattr(model,hookname)
        this_hook_dict = {k:hooks for k in this_subclass_dict}
        base_hooks.update(this_hook_dict)

    return  base_classes, base_hooks


def get_hook_dict(model,custom_hooks):

    return {classname:getattr(model,hookname)
            for classname, hookname in custom_hooks.items()}

def init_thermo_model_from_dict(new, obj):
    for rxn_dict in obj['reactions']:
        rxn = new.reactions.get_by_id(rxn_dict['id'])

        if 'thermo' in rxn_dict:
            rxn.thermo = rxn_dict['thermo']

    for met_dict in obj['metabolites']:
        met = new.metabolites.get_by_id(met_dict['id'])

        if 'thermo' in met_dict:
            new._prepare_metabolite(met)
            met.thermo = MetaboliteThermo(metData=None, pH=7,ionicStr=0)
            for k,v in met_dict['thermo'].items():
                setattr(met.thermo,k,v)
    return new

def rebuild_compositions(new, compositions_dict):
    """
    Performs the reverse operation of :func:archive_compositions

    :param new:
    :param compositions_dict:
    :return:
    """

    return {k:_rebuild_stoichiometry(new,stoich)
            for k,stoich in compositions_dict.items()}

def _rebuild_stoichiometry(new, stoich):
    """
    Performs the reverse operation of :func:_stoichiometry_to_dict

    :param new:
    :param stoich:
    :return:
    """

    return defaultdict(int,
                       {new.metabolites.get_by_id(k):v
                        for k,v in stoich.items()})



"""
.. module:: pytfa
   :platform: Unix, Windows
   :synopsis: Thermodynamics-based Flux Analysis

.. moduleauthor:: pyTFA team

Tools to import or export enrichment to and from pytfa models


"""
import json

import pandas as pd

from .json import MyEncoder


def write_lexicon(tmodel, filepath):
    """
    Writes a csv file in the format :
                    seed_id
    13BDglcn_c    cpd11791
    13dpg_c       cpd00203
    2pg_c         cpd00482
    3pg_c         cpd00169
    4abut_c       cpd00281

    Useful for exporting an annotation

    :type tmodel: pytfa.core.ThermoModel
    :param tmodel:
    :param filepath:
    :return:
    """

    lexicon = pd.DataFrame.from_dict({x.id:x.annotation    \
                                   for x in tmodel.metabolites},
                                  orient = 'index')
    lexicon.to_csv(filepath)
    return lexicon


def annotate_from_lexicon(model,lexicon):
    """
    Converts a lexicon into annotation for the metabolites

    :type model: cobra.Model
    :param model:
    :param lexicon:
    :return:
    """
    annotations = lexicon.to_dict(orient = 'index')

    for this_metabolite in model.metabolites:
        try:
            this_metabolite.annotation = annotations[this_metabolite.id]
        except KeyError:
            model.logger.warning(this_metabolite.id +   \
                                 '  not found in annotations')


def read_lexicon(filepath):
    return pd.read_csv(filepath, index_col = 0)


def write_compartment_data(tmodel, filepath):
    """

    :param filepath:
    :type tmodel: pytfa.core.ThermoModel
    :param tmodel:
    :return:
    """

    # compartment_data = pd.DataFrame.from_dict(tmodel.compartments,
    #                                           orient = 'index')
    # compartment_data.to_csv(filepath)
    # return compartment_data
    filepath = filepath + '.json' if not filepath.endswith('.json') \
            else filepath
    with open(filepath, 'w') as outfile:
        json.dump(tmodel.compartments, outfile, cls=MyEncoder)


def read_compartment_data(filepath):
    #return pd.read_csv(filepath, index_col = 0)
    filepath = filepath + '.json' if not filepath.endswith('.json') \
            else filepath
    with open(filepath) as json_data:
        compartment_data = json.load(json_data)
        return compartment_data


def apply_compartment_data(tmodel,compartment_data):
    # tmodel.compartments = compartment_data.to_dict(orient='index')
    tmodel.compartments = compartment_data





# -*- coding: utf-8 -*-
"""
.. module:: pytfa
   :platform: Unix, Windows
   :synopsis: Thermodynamics-based Flux Analysis

.. moduleauthor:: pyTFA team

Plotting results

"""
import numpy as np
import pandas as pd

from bokeh.models import ColumnDataSource, DataRange1d, LinearAxis, \
    Grid,CategoricalTicker, FuncTickFormatter
from bokeh.models.glyphs import HBar
from bokeh.io import show
from bokeh.plotting import figure

from ..optim.variables import ThermoDisplacement

def plot_fva_tva_comparison(fva,tva):

    all_va = pd.merge(fva,tva, left_index = True, right_index=True)

    all_va['y'] = range(len(all_va))

    source = ColumnDataSource(all_va)

    xdr = DataRange1d()
    ydr = DataRange1d()

    plot = figure(title=None, x_range=xdr, y_range=ydr, plot_width=600,
        plot_height=1000, h_symmetry=False, v_symmetry=False, min_border=0)

    glyph_fva = HBar(y="y", right="maximum_x", left="minimum_x", height=0.5,
                  fill_color="#b3de69", fill_alpha = 0.5)
    glyph_tva = HBar(y="y", right="maximum_y", left="minimum_y", height=0.8,
                  fill_color="#2e86c1", fill_alpha = 0.5)
    plot.add_glyph(source, glyph_tva)
    plot.add_glyph(source, glyph_fva)


    # Fix ticks
    label_dict = {}
    for i, s in enumerate(tva.index):
        label_dict[i] = s
    plot.yaxis.formatter = FuncTickFormatter(code="""
                                            var labels = %s;
                                            return labels[tick];
                                        """ % label_dict)

    plot.yaxis.ticker = [x for x in range(len(tva))]

    return plot

def plot_thermo_displacement_histogram(tmodel,solution = None):
    """
    Plot a histogram of the thermodynamic displacement. if no solution is
    provided, will look at the cobra_model's own solution

    :param tmodel:
    :param solution:
    :return:
    """

    if solution is None:
        values = tmodel.get_primal(ThermoDisplacement)
    else:
        thermo_displacements = tmodel.get_variables_of_type(ThermoDisplacement)
        varnames = [x.name for x in thermo_displacements]

        values = solution.raw[varnames]

    p1 = plot_histogram(values)


    p1.legend.location = "center_right"
    p1.legend.background_fill_color = "darkgrey"
    p1.xaxis.axis_label = 'ln(Γ)'
    p1.yaxis.axis_label = 'count'
    p1.title = "Histogram: Thermodynamic displacement"

    return p1

def plot_histogram(values, **kwargs):
    """
    Convenience function. Plots a histogram of flat 1D data.

    :param values:
    :return:
    """


    hist, edges = np.histogram(values, **kwargs)

    p1 = figure(tools="save")

    p1.quad(top=hist, bottom=0, left=edges[:-1], right=edges[1:],
            fill_color="#036564", line_color="#033649")


    return p1


# -*- coding: utf-8 -*-
"""
.. module:: pytfa
   :platform: Unix, Windows
   :synopsis: Thermodynamics-based Flux Analysis

.. moduleauthor:: pyTFA team

Variability analysis

"""

from copy import deepcopy
from functools import partial
from multiprocessing import cpu_count
from multiprocessing.pool import Pool

import pandas as pd
from cobra.core import Reaction
from optlang.exceptions import SolverError
from optlang.interface import INFEASIBLE
from tqdm import tqdm

from ..optim import DeltaG
from ..optim.constraints import ForbiddenProfile
from ..optim.utils import get_direction_use_variables
from ..optim.variables import ForwardUseVariable
from ..utils.logger import get_bistream_logger

CPU_COUNT = cpu_count()
BEST_THREAD_RATIO = int(CPU_COUNT/(4*2))    # Four proc per MILP instance,
                                            # times two threads.

def find_bidirectional_reactions(va, tolerance = 1e-8):
    """
    Returns the ids of reactions that can both carry net flux in the forward or
    backward direction.

    :param va: A variability analysis, pandas Dataframe like so:
                                          maximum       minimum
                6PGLter             -8.330667e-04 -8.330667e-04
                ABUTt2r              0.000000e+00  0.000000e+00
                ACALDt               0.000000e+00  0.000000e+00

    :return:
    """

    return va[va['minimum']*va['maximum'] < -tolerance]


def find_directionality_profiles(tmodel, bidirectional, max_iter = 1e4,
                                 solver = 'optlang-glpk'):
    """
    Takes a ThermoModel and performs enumeration of the directionality profiles

    :param tmodel:
    :param max_iter:
    :return:
    """

    raise(NotImplementedError)

    this_tmodel = deepcopy(tmodel)
    this_tmodel.solver = solver
    profiles = dict()

    iter_count = 0

    bidirectional_reactions = this_tmodel.reactions.get_by_any(bidirectional)

    while this_tmodel.solver.status != INFEASIBLE and iter_count < max_iter:

        try:
            solution = this_tmodel.optimize()
        except SolverError:
            break

        profiles[iter_count] = solution.raw
        if iter_count > 0:
            sse = sum((profiles[iter_count-1] - profiles[iter_count])**2)
        else:
            sse =0

        tmodel.logger.debug(str(iter_count) + ' - ' + str(sse))

        # active_use_variables = get_active_use_variables(this_tmodel,solution)
        active_use_variables = get_direction_use_variables(this_tmodel,solution)
        bidirectional_use_variables = [x for x in active_use_variables \
                                       if x.reaction in bidirectional_reactions]
        bool_id = ''.join('1' if isinstance(x,ForwardUseVariable) else '0' \
                for x in bidirectional_use_variables)

        # Make the expression to forbid this expression profile to happen again
        # FP_1101: FU_rxn1 + FU_rxn2 + BU_rxn3 + FU_rxn4 <= 4-1 = 3
        expr = sum(bidirectional_use_variables)
        this_tmodel.add_constraint(ForbiddenProfile,
                                   hook = this_tmodel,
                                   expr = expr,
                                   id = str(iter_count) + '_' + bool_id,
                                   lb = 0,
                                   ub = len(bidirectional_use_variables)-1)

        iter_count += 1

    return profiles,this_tmodel



def _bool2str(bool_list):
    """
    turns a list of booleans into a string

    :param bool_list: ex: '[False  True False False  True]'
    :return: '01001'
    """
    return ''.join(['1' if x else '0' for x in bool_list])


def _variability_analysis_element(tmodel, var, sense):
    tmodel.objective = var
    tmodel.objective.direction = sense
    sol = tmodel.slim_optimize()
    return sol



def variability_analysis(tmodel, kind='reactions', proc_num = BEST_THREAD_RATIO):
    """
    Performs variability analysis, gicven a variable type

    :param tmodel:
    :param kind:
    :param proc_num:
    :return:
    """

    objective = tmodel.objective

    # If the kind variable is iterable, we perform variability analysis on each,
    # one at a time
    if hasattr(kind, '__iter__') and not isinstance(kind, str):
        va = {}
        for k in kind:
            va[k] = variability_analysis(tmodel, kind=k, proc_num = proc_num)
        df = pd.concat(va.values())
        return df
    elif kind == Reaction or    \
            (isinstance(kind, str) and kind.lower() in ['reaction','reactions']):
        these_vars = {r.id : r for r in tmodel.reactions}
    else:
        these_vars = tmodel.get_variables_of_type(kind)
        these_vars = {x.name : x.variable for x in these_vars}

    tmodel.logger.info('Beginning variability analysis for variable of type {}'    \
                .format(kind))

    results = {'min':{}, 'max':{}}
    for sense in ['min','max']:
        for k,var in tqdm(these_vars.items(), desc=sense+'imizing'):
            tmodel.logger.debug(sense + '-' + k)
            results[sense][k] = _variability_analysis_element(tmodel,var,sense)


    tmodel.objective = objective
    df = pd.DataFrame(results)
    df.rename(columns={'min':'minimum','max':'maximum'}, inplace = True)
    return df


def parallel_variability_analysis(tmodel, kind='reactions', proc_num = BEST_THREAD_RATIO):
    """
    WIP.

    :param tmodel:
    :param kind:
    :param proc_num:
    :return:
    """

    raise(NotImplementedError)

    objective = tmodel.objective

    if kind == Reaction or kind.lower() in ['reaction','reactions']:
        these_vars = tmodel.reactions
    else:
        these_vars = tmodel.get_variables_of_type(kind)

    func = partial(_variability_analysis_element, tmodel)

    pool = Pool(processes=proc_num)
    async_result = pool.map_async(func, these_vars)
    pool.close()
    pool.join()

    # aggregated_result = pd.DataFrame(async_result.get(),
    #                                  columns = ['minimize','maximize'])

    tmodel.objective = objective
    return async_result


def calculate_dissipation(tmodel,solution=None):
    if solution is None:
        solution = tmodel.solution

    reaction_id  = [x.id for x in tmodel.reactions]
    fluxes = solution.fluxes[reaction_id]

    deltag_var = tmodel.get_variables_of_type(DeltaG)
    deltag = pd.Series({x.id:solution.raw[x.name]
                                      for x in deltag_var})
    dissipation = fluxes*deltag

    return dissipation


# -*- coding: utf-8 -*-
"""
.. module:: pytfa
   :platform: Unix, Windows
   :synopsis: Thermodynamics-based Flux Analysis

.. moduleauthor:: pyTFA team

MILP-fu to reformulate problems

"""
import sympy
# import optlang
from collections import namedtuple
from .variables import LinearizationVariable
from .constraints import LinearizationConstraint

# Faster than optlang Constraint object
ConstraintTuple = namedtuple('ConstraintTuple',['name','expression','ub','lb'])

OPTLANG_BINARY = 'binary'

def subs_bilinear(expr):
    """
    Substitutes bilinear forms from an expression with dedicated variables
    :param expr:
    :return:
    """

    bilinear_ix = [isinstance(x,sympy.Mul) for e,x in enumerate(expr.args)]

    new_expr = expr.copy()

    replacement_dict = dict()

    for bix in bilinear_ix:
        term = expr.args[bix]
        name = '__MUL__'.join(term.args)
        z = sympy.Symbol(name = name)

        new_expr = new_expr.subs(term,z)
        replacement_dict[term] = z

    return new_expr, replacement_dict


def glovers_linearization(b, fy, z = None, L=0, U=1000):
    """
    Glover, Fred.
    "Improved linear integer programming formulations of nonlinear integer problems."
    Management Science 22.4 (1975): 455-460.

    Performs Glovers Linearization of a product
    z = b*f(y) <=> z - b*f(y) = 0
    <=>
    {   L*b <= z <= U*b
    {   f(y) - U*(1-b) <= z <= f(y) - L*(1-b)

    where :
    * b is a binary variable
    * f a linear combination of continuous or integer variables y

    :param b:   Must be a binary optlang variable
    :param z:   Must be an optlang variable. Will be mapped to the product so
                that z = b*f(y)
    :param fy:  Must be an expression or variable
    :param L:   minimal value for fy
    :param U:   maximal value for fy
    :return:
    """

    assert(b.type == OPTLANG_BINARY)

    if z is None:
        name = '__MUL__'.join([b.name, fy.name])
        z = sympy.Symbol(name = name)

    # 1st Glovers constraint
    # L*b <= z
    # 0 <= z - L*b
    cons1 = optlang.Constraint(name = name + '_1',
                               expression = z - L*b,
                               lb = 0)
    # 2nd Glovers constraint
    # z <= U*b
    # 0 <= U*b - z
    cons2 = optlang.Constraint(name = name + '_2',
                               expression = U*b - z,
                               lb=0)

    # 3rd Glovers constraint
    # fy - U*(1-b) <= z
    # 0 <= z - fy + U*(1-b)
    cons3 = optlang.Constraint(name = name + '_3',
                               expression = z - fy + U*(1-b),
                               lb = 0)
    # 4th Glovers constraint
    # z <= fy - L*(1-b)
    # 0 <= fy - L*(1-b) - z
    cons4 = optlang.Constraint(name = name + '_4',
                               expression = fy - L*(1-b) - z,
                               lb=0)

    return z, [cons1,cons2,cons3,cons4]


def petersen_linearization(b, x, z = None, M=1000):
    """
    PETERSEN, C,,
    "A Note on Transforming the Product of Variables to Linear Form in Linear CLIFFORD Programs,"
    Working Paper, Purdue University, 1971.

    Performs Petersen Linearization of a product
    z = b*x <=> z - b*x = 0
    <=>
    {   x + M*b - M <= z <= M*b
    {   z <= x

    where :
    * b is a binary variable
    * f a linear combination of continuous or integer variables y

    :param x:   Must be an expression or variable
    :param b:   Must be a binary optlang variable
    :param z:   Must be an optlang variable. Will be mapped to the product so
                that z = b*f(y)
    :param M:   big-M constraint
    :return:
    """

    assert(b.type == OPTLANG_BINARY)

    if z is None:
        name = '__MUL__'.join([b.name, x.name])
        z = sympy.Symbol(name = name)
    else:
        name = z.name

    # 1st Petersen constraint
    # x + M*b - M <= z
    # x + M*b - z <= M
    cons1 = ConstraintTuple(name = name + '_1',
                            expression = x + M*b - z,
                            lb=0,
                            ub = M)
    # 2nd Petersen constraint
    # z <= M*b
    # 0 <= M*b - z
    cons2 = ConstraintTuple(name = name + '_2',
                            expression = M*b - z,
                            lb=0,
                            ub=None)

    # 3rd Petersen constraint
    # z <= x
    # 0 <= x - z
    cons3 = ConstraintTuple(name = name + '_3',
                            expression = x - z,
                            lb = 0,
                            ub = None,
                            )

    return z, [cons1,cons2,cons3]


def linearize_product(model, b, x, queue=False):
    """

    :param model:
    :param b: the binary variable
    :param x: the continuous variable
    :param queue: whether to queue the variables and constraints made
    :return:
    """

    # Linearization step for ga_i * [E]
    z_name = '__MUL__'.join([b.name, x.name])
    # Add the variables
    model_z_u = model.add_variable(kind=LinearizationVariable,
                                  hook=model,
                                  id_=z_name,
                                  lb=0,
                                  ub=x.ub,
                                  queue=False)

    big_m = x.ub

    z_u, new_constraints = petersen_linearization(b=b, x=x, M=big_m,
                                                  z=model_z_u)

    # Add the constraints:
    for cons in new_constraints:
        model.add_constraint(kind=LinearizationConstraint,
                             hook=model,
                             id_=cons.name,
                             expr=cons.expression,
                             # expr=new_expression,
                             ub=cons.ub,
                             lb=cons.lb,
                             queue=queue)

    model._push_queue()
    return model_z_u




#!/usr/bin/env python

"""
Thermo Database Converter, Matlab to Python

.. module:: pyTFA
   :platform: Unix, Windows
   :synopsis: Thermodynamics-based Flux Analysis

.. moduleauthor:: pyTFA team

:date: June 13th, 2017
:author: pyTFA team

Usage: python thermoDBconverter.py source_database.mat destination.thermodb

You may also import matToPy in your script to convert a thermoDB from matlab to
a python-friendly equivalent of it.


"""


import sys
import os
import numpy
import re
import pickle
import zlib
from scipy.io import loadmat


def matToPy(db):
    """ Convert a Matlab-imported array to a Python Database

    :params numpy.ndarray db: The array of an imported .mat file

    :returns: A dictionnary containing the database data, put in a
    python-friendly format.
    :rtype: dict

    """

    def parse_cue_data(value):
        """ Convert a struct_cue string to a Python dictionnary

        Args:
           value (str): The string to parse

        Returns:
           dict. A dictionnary with the parsed data from the string.

        This is just a little tool used while parsing cue data.
        """

        res = {}
        cues = re.split(r';|\|', value)

        for cue in cues:
            cue = cue.split(':')

            if len(cue) != 2: # Invalid data, for example NULL, NA,...
                continue # Just skip it

            res[cue[0]] = int(cue[1])

        return res

    # Which unit are we using here ?
    units = db['thermo_units'][0,0][0]
    print("Units of the database : " + units)

    ###############
    # METABOLITES #
    ###############

    CompoundDB = db['compound'][0,0]

    # Number of metabolites in the database
    nb_compounds = len(CompoundDB['ID'][0,0])
    print(str(nb_compounds) + " metabolites in database")

    # Initialize the storage for our metabolites
    metabolites={}

    # Fill the data
    for i in range(nb_compounds):
        # Fill the data of the metabolite with the values from Matlab
        met = {}
        met['id'] = CompoundDB['ID'][0,0][i,0][0]
        met['mass_std'] = CompoundDB['Mass_std'][0,0][i,0]
        try:
            met['formula'] = CompoundDB['Formula'][0,0][i,0][0]
        except IndexError:
            met['formula'] = ''
        met['name'] = CompoundDB['metNames'][0,0][i,0][0]
        try:
            met['other_names'] = CompoundDB['AllNames'][0,0][i,0][0].split('|')
        except ValueError:
            met['other_names'] = []
        met['charge_std'] = CompoundDB['Charge_std'][0,0][i,0]

        # Make the cues' data a bit more friendly (as a dict)
        try:
            met['struct_cues'] = parse_cue_data(CompoundDB['struct_cues'][0,0][i,0][0])
        except IndexError:
            met['struct_cues'] = []

        # Continue the import...
        met['nH_std'] = CompoundDB['nH_std'][0,0][i,0]
        met['deltaGf_std'] = CompoundDB['deltaGf_std'][0,0][i,0]
        met['deltaGf_err'] = CompoundDB['deltaGf_err'][0,0][i,0]
        met['error'] = CompoundDB['error'][0,0][i,0][0]
        met['nH_std'] = CompoundDB['nH_std'][0,0][i,0]

        # Special treatment for pkA : convert it to a list of floating values
        met['pKa'] = CompoundDB['pKa'][0,0][i,0][0]
        # BUG: Somehow, the conversion fails for some values and we have sub-arrays...
        if type(met['pKa']) == numpy.ndarray:
            met['pKa'] = met['pKa'][0][0]

        # Create the list...
        met['pKa'] = met['pKa'].split('|')
        # Remove empty values
        met['pKa'] = [x for x in met['pKa'] if x not in [None, '']]
        # If there aren't any values, remove the list
        if met['pKa'][0] == 'NA':
            met['pKa'] = []
        else:
            # Convert the values to float (they're strings)
            met['pKa'] = [float(pKa) for pKa in met['pKa']]

        # Register the metabolite for easier access
        if met['id'] in metabolites:
            raise Exception("Duplicate ID : " + met['id'])
        metabolites[met['id']] = met

    ########
    # CUES #
    ########
    cues = {}

    CueDB = db['cue'][0,0]
    nb_cue = len(CueDB['ID'][0,0])

    for i in range(nb_cue):
        # Import the cue's data
        cue = {}
        cue['id'] = CueDB['ID'][0,0][i,0][0]
        cue['error'] = CueDB['Error'][0,0][i,0][0,0]
        cue['charge'] = CueDB['Charge'][0,0][i,0][0,0]
        cue['energy'] = CueDB['Energy'][0,0][i,0][0,0]
        # If the value is empty, set it to None
        try:
            cue['datfile'] = CueDB['Datfile'][0,0][i,0]
            cue['datfile'] = cue['datfile'][0] if cue['datfile'] else None
        except ValueError:
            cue['datfile'] = None
        cue['formula'] = CueDB['Formula'][0,0][i,0]
        cue['formula'] = cue['formula'][0] if cue['formula'] else None
        try:
            cue['small']   = True if CueDB['Small_Molecule'][0,0][i,0][0,0] == 1 else False
        except IndexError:
            cue['small'] = False
        cue['names']   = CueDB['AllNames'][0,0][i,0][0].split('|')

        # Register the cue in the database
        cues[cue['id']] = cue

    return (units, metabolites, cues)


# If the script was called directly (not imported)
if __name__ == "__main__":
    # Make sure we have all the parameters...
    if len(sys.argv) < 3:
        print("Usage : python " + __file__ + " source_database.mat destination.thermodb")
        sys.exit(1)

    # Display a warning if the file already exists
    if os.path.isfile(sys.argv[2]):
        print("WARNING: The destination file already exists, it will be overwritten")

    # Load the Matlab file...
    print("Loading the database in memory...")

    try:
        ReactionDB = loadmat(sys.argv[1])
    except:
        print("Cannot load source file")
        sys.exit(2)

    # Guess the key of the array to the data of the database
    print("Guessing the name of the database...")
    name = None
    for key in ReactionDB:
        if key[:2] == "__": # Name begins with "__" (private attribute)
            continue
        if name == None:
            name = key
        else:
            print("Cannot guess the name of the database ! Found at least 2 candidates : ")
            print(name + " and " + key)
            sys.exit(3)

    if name == None:
        print("Cannot find name, no candidates found")
        sys.exit(3)

    print("Found name : " + name)
    ReactionDB = ReactionDB[name]

    # Convert all the matlab data to a more python-friendly version
    (units, metabolites, cues) = matToPy(ReactionDB)

    # Finally, save the data
    try:
        with open(sys.argv[2], 'wb') as file:
            # Convert it to string with Pickle
            data = pickle.dumps({
                "name":name,
                "units": units,
                "metabolites":metabolites,
                "cues":cues
            }, pickle.HIGHEST_PROTOCOL)

            # Compress it with zlib
            data = zlib.compress(data)

            # And write it to disk
            file.write(data)
    except:
        print('Could not save the data')
        sys.exit(4)

    sys.exit(0)



# -*- coding: utf-8 -*-
"""
.. module:: pytfa
   :platform: Unix, Windows
   :synopsis: Thermodynamics-based Flux Analysis

.. moduleauthor:: pyTFA team

JSON serialization
"""

import json

import numpy

from .dict import model_from_dict, model_to_dict


class MyEncoder(json.JSONEncoder):
    """
    We define an encoder that takes care of the serialization of numpy types,
    which are not handled by json by default
    """
    def default(self, obj):
        if isinstance(obj, numpy.integer):
            return int(obj)
        elif isinstance(obj, numpy.floating):
            return float(obj)
        elif isinstance(obj, numpy.ndarray):
            return obj.tolist()
        else:
            return super(MyEncoder, self).default(obj)


def check_json_extension(filepath):
    if not filepath.endswith('.json'):
        filepath += '.json'
    return filepath

def save_json_model(model, filepath):

    filepath = check_json_extension(filepath)
    obj = model_to_dict(model)

    with open(filepath, 'w') as fid:
        json.dump(obj, fid, cls=MyEncoder)


def load_json_model(filepath):

    filepath = check_json_extension(filepath)
    with open(filepath, 'r') as fid:
        obj = json.load(fid)

    model = model_from_dict(obj)
    return model


def json_dumps_model(model):
    """
    Returns a JSON dump as a string

    :param model:
    :return:
    """

    obj = model_to_dict(model)

    return json.dumps(obj,cls=MyEncoder)


def json_loads_model(s):
    """
    Loads a model from a string JSON dump

    :param s: JSON string
    :return:
    """
    obj = json.loads(s)
    return model_from_dict(obj)


#!/usr/bin/env python

# -*- coding: utf-8 -*-

"""
.. module:: redgem
   :platform: Unix, Windows
   :synopsis: RedGEM Algorithm

.. moduleauthor:: pyTFA team

Debugging
"""

from cobra import Reaction
from pandas import Series

def make_sink(met, ub=100, lb=0):
    rid = 'sink_' + met.id
    try:
        # if the sink already exists 
        # (happens when debugging several times a model)
        new = met.model.reactions.get_by_id(rid)
    except KeyError:
        new = Reaction(id = rid)
        new.add_metabolites({met:-1})
    
    new.lower_bound = lb
    new.upper_bound = ub

    return new

def add_BBB_sinks(model,biomass_rxn_id, ub=100, lb=0):

    bio_rxn = model.reactions.get_by_id(biomass_rxn_id)

    all_BBBs = bio_rxn.reactants
    all_sinks = list()

    for the_BBB in all_BBBs:
        new = make_sink(the_BBB, ub=ub, lb=lb)
        all_sinks.append(new)

    model.add_reactions([x for x in all_sinks if not x.id in model.reactions])
    return [model.reactions.get_by_id(x.id) for x in all_sinks]

def check_BBB_production(model, biomass_rxn_id, verbose = False):

    all_sinks = add_BBB_sinks(model, biomass_rxn_id, lb = 0)

    prod = dict()

    for the_sink in all_sinks:
        with model:
            model.objective = the_sink.id
            prod[the_sink.id] = model.slim_optimize()

    ret = Series(prod)
    if verbose:
        print(ret)
    return ret

def min_BBB_uptake(model,biomass_rxn_id, min_growth_value, verbose=False):
    

    with model:
        all_sinks = add_BBB_sinks(model, biomass_rxn_id, ub = 0, lb = -100)
        # Uptake is negative
        # Min absolute uptake = Max uptake
        bio_rxn = model.reactions.get_by_id(biomass_rxn_id)
        bio_rxn.lower_bound = min_growth_value
        model.objective = sum( - 1* s.reverse_variable for s in all_sinks)
        model.objective_direction = 'max'
        model.optimize()
        ret = Series({r:r.flux for r in all_sinks})

    if verbose:
        print(ret)
    return ret


from ..core.model import Solution
import pandas as pd

def apply_reaction_variability(tmodel, va, inplace = True):
    """
    Applies the VA results as bounds for the reactions of a cobra_model
    :param inplace:
    :param tmodel:
    :param va:
    :return:
    """
    if inplace:
        _tmodel = tmodel
    else:
        _tmodel = tmodel.copy()

    for this_reaction in _tmodel.reactions:
        if this_reaction.id not in va.index:
            continue

        this_reaction.lower_bound = va.loc[this_reaction.id,'minimum']
        this_reaction.upper_bound = va.loc[this_reaction.id,'maximum']

    return _tmodel


def apply_generic_variability(tmodel,va, inplace = True):
    """
    Reactions a dealt with cobra, but the other variables added use pytfa's
    interface: the class GenericVariable. We use a different method to apply
    variability directly in the solver

    :param tmodel:
    :param va:
    :param inplace:
    :return:
    """
    if inplace:
        _tmodel = tmodel
    else:
        _tmodel = tmodel.copy()


    for varname in va.index:
        the_min,the_max = va.loc[varname,['minimum','maximum']]
        _tmodel._var_dict[varname].variable.lb = the_min
        _tmodel._var_dict[varname].variable.ub = the_max

    return _tmodel


def apply_directionality(tmodel, solution, inplace = True):
    """
    Takes a flux solution and transfers its reaction directionality as
    constraints for the cobra_model

    :param inplace:
    :param tmodel:
    :param solution:
    :return:
    """

    if inplace:
        _tmodel = tmodel
    else:
        _tmodel = tmodel.copy()

    if isinstance(solution, Solution):
        sol = solution.raw
    elif isinstance(solution, pd.Series) or isinstance(solution, pd.DataFrame):
        sol = solution
    else:
        raise ArgumentError('solution object should be of class Solution or pandas.Series')

    for this_reaction in _tmodel.reactions:

        backward_use = _tmodel.backward_use_variable.get_by_id(this_reaction.id)
        forward_use = _tmodel.forward_use_variable.get_by_id(this_reaction.id)

        backward_use.variable.lb = round(sol[backward_use.name])
        backward_use.variable.ub = round(sol[backward_use.name])

        forward_use.variable.lb  = round(sol[forward_use.name])
        forward_use.variable.ub  = round(sol[forward_use.name])

    return _tmodel



# -*- coding: utf-8 -*-
"""
.. module:: pytfa
   :platform: Unix, Windows
   :synopsis: Thermodynamics-based Flux Analysis

.. moduleauthor:: pyTFA team

Metaclass declarations to force the definition of prefixes in GenericVariable
and GeneriConstraint subclasses

Based on SethMMorton's answer on StackOverflow
https://stackoverflow.com/questions/45248243/most-pythonic-way-to-declare-an-abstract-class-property
https://stackoverflow.com/a/45250114
"""
from abc import ABCMeta

class RequirePrefixMeta(type):
    """Metaclass that enforces child classes define prefix."""

    def __init__(cls, name, bases, attrs):
        # Skip the check if there are no parent classes,
        # which allows base classes to not define prefix.
        if not bases:
            return
        if attrs.get('prefix', NotImplemented) is NotImplemented:
            # Choose your favorite exception.
            raise NotImplementedError('You forgot to define a prefix for your class ;)')

ABCRequirePrefixMeta = type('ABCRequirePrefixMeta',
                            (ABCMeta, RequirePrefixMeta), {})



# -*- coding: utf-8 -*-
"""
.. module:: pytfa
   :platform: Unix, Windows
   :synopsis: Thermodynamics-based Flux Analysis

.. moduleauthor:: pyTFA team

Debugging of models

"""

from collections import defaultdict
import pandas as pd

def debug_iis(model):
    """
    Performs reduction to an Irreducible Inconsistent Subsystem (IIS)

    :param model:
    :return:
    """
    out_c = list()
    out_v = list()
    if model.solver.__class__.__module__ == 'optlang.gurobi_interface':
        model.solver.problem.computeIIS()
        print('# Constraints:')
        for c in model.solver.problem.getConstrs():
            if c.IISConstr:
                out_c.append(c)
                print('%s' % c.constrName)
        print('# Variables:')
        for v in model.solver.problem.getVars():
            if v.IISUB != 0 or v.IISLB != 0:
                out_v.append(v)
                print('{}: IISLB = {}, IISUB = {}, (original bounds {}, {})'\
                      .format(v.VarName, v.IISLB, v.IISUB, v.LB, v.UB))
        # model.write("IIS_debug_{}.ilp".format(model.name))
    elif model.solver.__class__.__module__ == 'optlang.cplex_interface':
        model.solver.problem.conflict.refine(
                model.solver.problem.conflict.all_constraints())
    else:
        model.logger.error('Not implemented for solver {}'.format(model.solver))

    return out_c, out_v

def find_extreme_coeffs(model,n=5):
    max_coeff_dict = defaultdict(int)
    min_coeff_dict = defaultdict(lambda:1000)
    max_cons_dict = dict()
    min_cons_dict = dict()

    for the_cons in model.constraints:
        for the_var, the_coeff in the_cons.expression.as_coefficients_dict().items():
            if abs(the_coeff) > max_coeff_dict[the_var.name]:
                max_coeff_dict[the_var.name] = abs(the_coeff)
                max_cons_dict[the_var.name] = the_cons.name
            if 0 < abs(the_coeff) < min_coeff_dict[the_var.name]:
                min_coeff_dict[the_var.name] = abs(the_coeff)
                min_cons_dict[the_var.name] = the_cons.name

    def prep_result(cons_dict, coeff_dict):
        coeff_data = pd.DataFrame.from_dict(coeff_dict, orient = 'index')
        cons_data = pd.DataFrame.from_dict(cons_dict, orient = 'index')

        res = pd.concat([cons_data, coeff_data], axis = 1)
        res.columns = ['constraint','coeff']
        res.index.name = 'variable'
        return res


    ret1 = prep_result(max_cons_dict, max_coeff_dict)\
        .sort_values('coeff',ascending=False).head(n)
    ret2 = prep_result(min_cons_dict, min_coeff_dict)\
        .sort_values('coeff',ascending=True).head(n)

    return pd.concat([ret1, ret2], axis = 0)

def find_maxed_vars(model, ub = 1000, epsilon = 1e-2):
    ret = [(x.name, x.primal, x.ub)
            for x in model.variables
            if abs(x.ub - x.primal) <epsilon and ub>x.ub >0
                                            and not x.type=='binary']

    return pd.DataFrame(ret)


# -*- coding: utf-8 -*-
"""
.. module:: pytfa
   :platform: Unix, Windows
   :synopsis: Thermodynamics-based Flux Analysis

.. moduleauthor:: pyTFA team

Relaxation of models with constraint too tight

"""
from copy import deepcopy

import optlang
import pandas as pd
import sympy
from cobra.core.solution import Solution

from numbers import Number

from .constraints import GenericConstraint
from .variables import ForwardUseVariable, BackwardUseVariable
from .variables import GenericVariable

SYMPY_ADD_CHUNKSIZE = 100
INTEGER_VARIABLE_TYPES = ('binary','integer')

def get_all_subclasses(cls):
    """
    Given a variable or constraint class, get all the subclassses
    that inherit from it

    :param cls:
    :return:
    """
    all_subclasses = []

    for subclass in cls.__subclasses__():
        all_subclasses.append(subclass)
        all_subclasses.extend(get_all_subclasses(subclass))

    return all_subclasses

def chunk_sum(variables):
    """
    This functions handles the sum of many sympy variables by chunks, which
    somehow increases the speed of the computation

    You can test it in IPython:
    ```python
    a = sympy.symbols('a0:100')
    %timeit (sum(a))
    # >>> 198 µs ± 11.4 µs per loop (mean ± std. dev. of 7 runs, 1 loop each)

    b = sympy.symbols('b0:1000')
    %timeit (sum(b))
    # >>> 1.85 ms ± 356 µs per loop (mean ± std. dev. of 7 runs, 1 loop each)

    c = sympy.symbols('c0:3000')
    %timeit (sum(c))
    # >>> 5min 7s ± 2.57 s per loop (mean ± std. dev. of 7 runs, 1 loop each)
    ```

    See the `github thread <https://github.com/sympy/sympy/issues/13945>`_

    :param variables:
    :return:
    """

    partial_sums = []

    for chunk_no in range(len(variables) % SYMPY_ADD_CHUNKSIZE):
        first_index = chunk_no * SYMPY_ADD_CHUNKSIZE
        last_index = (chunk_no + 1) * SYMPY_ADD_CHUNKSIZE
        this_chunk = variables[first_index:last_index]
        this_sum = sum(this_chunk)
        partial_sums.append(this_sum)

    return sum(partial_sums)

def symbol_sum(variables):
    """
    
    ``` python
    a = symbols('a0:100')
    
    %timeit Add(*a)
    # >>> 10000 loops, best of 3: 34.1 µs per loop
    
    b = symbols('b0:1000')
    
    %timeit Add(*b)
    # >>> 1000 loops, best of 3: 343 µs per loop
    
    c = symbols('c0:3000')
    
    %timeit Add(*c)
    # >>> 1 loops, best of 3: 1.03 ms per loop
    ```
    
    See the `github thread <https://github.com/sympy/sympy/issues/13945>`_
    :param variables:
    :return:
    """

    from sympy import Add
    
    k=0
    # If we encounter a zero, which is a special type, increase k
    while isinstance(variables[k], sympy.numbers.Zero) and k<len(variables):
        k+=1
        if k == len(variables):
            # everything is 0
            return 0

    if k>len(variables): #it's only zeroes
        return 0

    if isinstance(variables[k], GenericVariable):
        return Add(*[x.variable for x in variables])
    elif isinstance(variables[k], optlang.interface.Variable) or    \
         isinstance(variables[k], sympy.Mul) or \
         isinstance(variables[k], sympy.Add) or \
         isinstance(variables[k], Number):
        return Add(*variables)
    else:
        raise ValueError('Arguments should be of type Number, sympy.Add, or sympy.Mul, '
                         'or optlang.Variable, or GenericVariable')


def get_solution_value_for_variables(solution, these_vars, index_by_reaction = False):
    if isinstance(these_vars[0],GenericVariable):
        var_ids = [x.name for x in these_vars]
    elif isinstance(these_vars[0],optlang.Variable):
        var_ids = [x.name for x in these_vars]
    elif isinstance(these_vars[0],str):
        var_ids = these_vars
    else:
        raise TypeError('''<these_vars> should be of type pytfa.solving.Variable,
                            optlang.Variable, or str''')


    if index_by_reaction:
        var2rxn = {v.name:v.id for v in these_vars}
        ret = solution.raw[var_ids]
        ret = ret.index.replace(var2rxn)
        return ret
    else:
        return solution.raw[var_ids]

def compare_solutions(models):
    """
    returns the solution dictionnary for each cobra_model
    :param (iterable (pytfa.thermo.ThermoModel)) models:
    :return:
    """
    return pd.concat([x.solution.raw for x in models], axis=1)

def evaluate_constraint_at_solution(constraint, solution):
    """

    :param expression:
    :param solution: pandas.DataFrame, with index as variable names
    :return:
    """

    if isinstance(solution,Solution):
        solution = solution.raw
    if isinstance(constraint, GenericConstraint):
        constraint = constraint.constraint

    # subs_dict = {x:solution.loc[x.name] for x in constraint.variables}
    # return constraint.expression.subs(subs_dict)

    coefs = constraint.get_linear_coefficients(constraint.expression.free_symbols)
    values = {x:solution.loc[x.name] for x in constraint.expression.free_symbols}

    return symbol_sum([coefs[x]*values[x] for x in coefs])


def get_active_use_variables(tmodel,solution):
    """
    Returns the active use variables in a solution. Use Variables are binary
    variables that control the directionality of the reaction
     ex:
     FU_ACALDt
     BU_PFK

    :type tmodel: pytfa.core.ThermoModel
    :param tmodel:
    :param solution:
    :return:
    """
    use_variables = tuple(tmodel.get_variables_of_type(BackwardUseVariable)) \
                  + tuple(tmodel.get_variables_of_type(ForwardUseVariable))

    epsilon = tmodel.solver.configuration.tolerances.integrality

    return [x for x in use_variables if abs(solution.raw[x.name]-1)<epsilon]


def get_direction_use_variables(tmodel,solution):
    """
    Returns the active use variables in a solution. Use Variables are binary
    variables that control the directionality of the reaction
    The difference with get_active_use_variables is that variables with both
    UseVariables at 0 will return as going forwards. This is to ensure that the
    output size of the function is equal to the number of FDPs
     ex:
     FU_ACALDt
     BU_PFK

    :type tmodel: pytfa.core.ThermoModel
    :param tmodel:
    :param solution:
    :return:
    """
    fwd_use_variables = tmodel.get_variables_of_type(ForwardUseVariable)
    bwd_use_variables = tmodel.get_variables_of_type(BackwardUseVariable)

    epsilon = tmodel.solver.configuration.tolerances.feasibility

    return [fwd_use_variables.get_by_id(x.id) if solution.raw[x.id] > epsilon
            else bwd_use_variables.get_by_id(x.id)
            for x in tmodel.reactions ]

def get_primal(tmodel, vartype, index_by_reactions = False):
    """
    Returns the primal value of the cobra_model for variables of a given type
    :param tmodel:
    :param vartype: Class of variable. Ex: pytfa.optim.variables.ThermoDisplacement
    :param index_by_reactions: Set to true to get reaction names as index instead of
        variables. Useful for Escher export
    :return:
    """

    the_vars = tmodel.get_variables_of_type(vartype)

    if index_by_reactions:
        return pd.Series({x.id:x.variable.primal
                          for x in the_vars})
    else:
        return pd.Series({x.name:x.variable.primal
                          for x in the_vars})


def strip_from_integer_variables(tmodel):
    """
    Removes all integer and binary variables of a cobra_model, to make it sample-able
    :param tmodel:
    :return:
    """
    continuous_model = tmodel.copy()
    continuous_model.name = tmodel.name + ' - continuous'

    integer_variables = set()

    constraints_with_integer_variables = []

    # We go through all the constraint descriptors and check if at least one of
    # their variables is in the integer variable list
    for this_cons in continuous_model._cons_dict.values():
        has_integer_variable = False
        for this_var in this_cons.constraint.variables:
            if this_var.type in INTEGER_VARIABLE_TYPES:
                has_integer_variable += True
                this_var_descriptor = continuous_model._var_dict[this_var.name]
                integer_variables.add(this_var_descriptor)
        if has_integer_variable:
            constraints_with_integer_variables.append(this_cons)

    for this_cons in constraints_with_integer_variables:
        continuous_model.remove_constraint(this_cons)

    for this_var in integer_variables:
        continuous_model.remove_variable(this_var)

    continuous_model.solver.update()
    # This will update the values =
    print('Is the cobra_model still integer ? {}'     \
          .format(continuous_model.solver.is_integer))

    return continuous_model

def copy_solver_configuration(source, target):
    """
    Copies the solver configuration from a source model to a target model
    :param source:
    :param target:
    :return:
    """

    # LP method
    try:
        target.solver.configuration.lp_method = source.solver.configuration.lp_method
    except AttributeError:
        pass

    # Presolve
    target.solver.configuration.presolve = source.solver.configuration.presolve

    # Timeout
    target.solver.configuration.timeout = source.solver.configuration.timeout

    # Tolerances
    for tol_name in dir(source.solver.configuration.tolerances):
        tol = getattr(source.solver.configuration.tolerances, tol_name)
        setattr(target.solver.configuration.tolerances, tol_name, tol)

    # Additionnal solver-specific settings
    try:
        # Gurobi
        if source.solver.interface.__name__ == 'optlang.gurobi_interface':
            from gurobipy import GurobiError
            for k in dir(source.solver.problem.Params):
                if not k.startswith('_'):
                    try:
                        v = getattr(source.solver.problem.Params, k)
                        setattr(target.solver.problem.Params, k, v)
                    except GurobiError:
                        pass
    except ModuleNotFoundError:
        pass

    # Verbosity
    target.solver.configuration.verbosity = source.solver.configuration.verbosity





"""
.. module:: pytfa
   :platform: Unix, Windows
   :synopsis: Thermodynamics-based Flux Analysis

.. moduleauthor:: pyTFA team

Input/Output tools to vizualize results


"""
import pandas as pd

from ..utils.str import varnames2ids


def export_variable_for_escher(tmodel,variable_type,data,filename):
    """
    Exports all the variables of a given type into a csv file, indexed by
    variable.id. This format is readable by escher if the variable_type is a
    subclass of :pytfa:`pytfa.optim.variables.ReactionVariable` or
    :pytfa:`pytfa.optim.variables.MetaboliteVariable`

    :param pytfa.core.ThermoModel tmodel:
    :param ReactionVariable|MetaboliteVariable variable_type:
    :param pandas.Series data: indexed by variable name
    :param string filename:
    :return:
    """
    var_dict = {x.name: x.id   \
                     for x in tmodel.get_variables_of_type(variable_type)}

    if isinstance(data,pd.DataFrame):
        # it is a analysis
        var_values = data[:, var_dict.keys()]
    elif isinstance(data,pd.Series):
        var_values = data[var_dict.keys()]

    var_values.index = varnames2ids(tmodel, var_values.index)
    var_values.to_csv(filename, header=True)


def get_reaction_data(tmodel, data):
    """
    Exports values indexed by reaction ids. Reconciles Forward and Backwards
    variables.
    """
    dict_values = {}
    for reaction in tmodel.reactions:
        fwd_val = data[reaction.forward_variable.name]
        bwd_val = data[reaction.reverse_variable.name]
        dict_values[reaction.id] = fwd_val - bwd_val

    var_values = pd.Series(dict_values)
    return  var_values


def export_reactions_for_escher(tmodel,data,filename):
    """
    Exports values indexed by reaction ids. Reconciles Forward and Backwards
    variables. Writes it in a csv file. This format is readable by escher

    :param pytfa.core.ThermoModel tmodel:
    :param ReactionVariable|MetaboliteVariable variable_type:
    :param pandas.Series data: indexed by variable name
    :param string filename:
    :return:
    """
    var_values = get_reaction_data(tmodel, data)
    var_values.to_csv(filename,header=True)


#!/usr/bin/env python
# -*- coding: utf-8 -*-

from cobra import Reaction

from ..optim.utils import symbol_sum
from ..thermo.utils import is_exchange, check_transport_reaction
from .utils import trim_epsilon_mets

from ..optim.variables import ReactionVariable, BinaryVariable, get_binary_type
from ..optim.constraints import ReactionConstraint, ForbiddenProfile

from numpy import sum, round

from optlang.interface import INFEASIBLE, TIME_LIMIT, OPTIMAL

from tqdm import tqdm

from collections import defaultdict, namedtuple

CPLEX = 'optlang-cplex'
GUROBI = 'optlang-gurobi'
GLPK = 'optlang-glpk'


# Transforms (OnePerBBB --> oneperbbb), (one_per_bbb --> oneperbbb), etc ...
disambiguate = lambda s:s.lower().replace('_','')

Lump = namedtuple('Lump', ['id_', 'metabolites', 'subnetwork', 'gene_reaction_rule'])

class InfeasibleExcept(Exception):
    def __init__(self, status, feasibility):
        self.status = status
        self.feasibility = feasibility


class TimeoutExcept(Exception):
    def __init__(self, time_limit):
        self.time_limit = time_limit


class FluxKO(ReactionVariable, BinaryVariable):
    prefix = 'KO_'

    def __init__(self, reaction, **kwargs):
        ReactionVariable.__init__(self, reaction,
                                  type=get_binary_type(),
                                  **kwargs)

# Define a new constraint type:
class UseOrKOInt(ReactionConstraint):
    prefix = 'UKI_'
# Define a new constraint type:
class UseOrKOFlux(ReactionConstraint):
    prefix = 'UKF_'


class LumpGEM:
    """
    A class encapsulating the LumpGEM algorithm
    """
    def __init__(self, tfa_model, additional_core_reactions, params):
        """
        :param tfa_model: The GEM (associated with the thermodynamics constraints) that lumpGEM must work on
        :type tfa_model: pytfa model

        :param biomass_rxns: list of biomass reactions
        :type biomass_rxns: [GEM.biomass_rxn.id]

        :param core_subsystems: list of Core subsystems names
        :type core_subsystems: [string]

        :param growth_rate: theoretical maximum specific growth rate in 1/hr units
        :type growth_rate: float

        :param timeout_limit: the maximum amount of time allowed to compute each optimization. Default is 3600s (1 hour)
        :type timeout_limit: float (seconds)
        """

        self._tfa_model = tfa_model

        self._param_dict = params
        self.init_params()

        # Set containing every BBB reaction
        self._rBBB = list()
        # Set containing every exchange reaction
        self._exchanges = list()
        # Set containing every transport reaction
        self._transports = list()
        # Set containing every core reaction
        self._rcore = list()
        # Set containing every non-core reaction
        self._rncore = list()

        # For each reaction
        for rxn in self._tfa_model.reactions:
            # If it's a BBB reaction
            if rxn.id in self.biomass_rxns:
                self._rBBB.append(rxn)
            # If it is an exchange reaction
            elif is_exchange(rxn):
                self._exchanges.append(rxn)
            # If it is a transport reaction
            elif check_transport_reaction(rxn):
                self._transports.append(rxn)
            # If it's a core reaction
            elif rxn.subsystem in self.core_subsystems:
                self._rcore.append(rxn)
            # If it is part of the intrasubsystem expansion
            elif rxn.id in additional_core_reactions:
                self._rcore.append(rxn)
            # If it's neither BBB nor core, then it's non-core
            else:
                self._rncore.append(rxn)

        # Growth rate
        self._growth_rate = self.growth_rate

        # TODO : solver choice
        # TODO default : solver du modele
        #self._solver = 'optlang-cplex'

        self._tfa_model.solver.configuration.timeout = self.timeout_limit
        print("Timeout limit is {}s".format(self.timeout_limit))

        # lumpgem binary variables to deactivate non-core reactions. The reaction is deactivated when the value of
        # the variable is 1
        self._activation_vars = {rxn: self._tfa_model.add_variable(kind=FluxKO,
                                                                   hook=rxn,
                                                                   lb=0,
                                                                   ub=1,
                                                                   queue=False)
                                 for rxn in self._rncore}

        self._generate_usage_constraints()
        self._generate_objective()
        self._sinks = self._prepare_sinks()

    def init_params(self):
        self.core_subsystems = self._param_dict["core_subsystems"]
        self.extracellular_system = self._param_dict["extracellular_system"]
        self.biomass_rxns = self._param_dict["biomass_rxns"]

        self.growth_rate = self._param_dict["growth_rate"]

        self.small_metabolites = self._param_dict["small_metabolites"]
        self.cofactor_pairs = self._param_dict["cofactor_pairs"]
        # Flatten cofactor_pairs list
        self.cofactors = [cofactor for pair in self.cofactor_pairs for cofactor in pair]
        self.inorganics = self._param_dict["inorganics"]

        self.timeout_limit = self._param_dict["timeout"]

        self.constraint_method = self._param_dict["constraint_method"]

    def _generate_usage_constraints(self):
        """
        Generate carbon intake related constraints for each non-core reaction
        For each reaction rxn : rxn.forward_variable + rxn.reverse_variable + activation_var * C_uptake < C_uptake
        """
        flux_methods = ['flux', 'fluxes', 'both']
        int_methods = ['int', 'integer', 'both']

        if self.constraint_method.lower() not in flux_methods + int_methods:
            raise ArgumentError('{} is not a correct constraint method. '
                                'Choose among [Flux, Integer, Both]. '
                                'If you do not know what to choose, go for Flux.'
                                'If it is too slow, go for integer.'
                                'If you get strange lumps, go for both'
                                .format(self.constraint_method))

        for rxn in self._rncore:
            activation_var = self._activation_vars[rxn]
            if self.constraint_method.lower() in flux_methods:
                bigM = 100
                reac_var = rxn.forward_variable + rxn.reverse_variable + activation_var * bigM
                # adding the constraint to the model
                self._tfa_model.add_constraint(kind=UseOrKOFlux,
                                               hook=rxn,
                                               expr=reac_var,
                                               ub=bigM,
                                               lb=0,
                                               queue=True)
            if self.constraint_method.lower() in int_methods:
                fu = self._tfa_model.forward_use_variable .get_by_id(rxn.id)
                bu = self._tfa_model.backward_use_variable.get_by_id(rxn.id)
                reac_var = fu + bu + activation_var
                # adding the constraint to the model
                self._tfa_model.add_constraint(kind=UseOrKOInt,
                                               hook=rxn,
                                               expr=reac_var,
                                               ub=1,
                                               lb=0,
                                               queue=True)


        # push constraints in one bulk (faster)
        self._tfa_model._push_queue()
        # refresh constraint fields
        self._tfa_model.repair()

    def get_cofactor_adjusted_stoich(self,rxn):
        stoich_dict = {x.id:v for x,v in rxn.metabolites.items()}

        for a,b in self.cofactor_pairs:
            try:
                na = stoich_dict[a] # looks like -54 atp_c
                nb = stoich_dict[b] # looks like +53 adp_c

                n = na+nb # looks like -1

                if n == 0:
                    self._tfa_model.logger.warn(
                        'Cofactor pair {}/{} is equimolar in reaction {}'
                        .format(a,b,rxn.id))
                elif n > 0:
                    n = -n
                    self._tfa_model.logger.warn(
                        'Cofactor pair {}/{} looks inverted in reaction {}'
                        .format(a,b,rxn.id))

                stoich_dict[a] =  n # looks like 1
                stoich_dict[b] = -n # looks like -1
            except KeyError:
                pass
        return stoich_dict


    def _prepare_sinks(self):
        """
        For each BBB (reactant of the biomass reactions), generate a sink, i.e an unbalanced reaction BBB ->
        of which purpose is to enable the BBB to be output of the GEM
        :return: the dict {BBB: sink} containing every BBB (keys) and their associated sinks
        """
        all_sinks = {}
        print("Preparing sinks...")

        for bio_rxn in self._rBBB:
            stoich_dict = self.get_cofactor_adjusted_stoich(bio_rxn)
            for met in bio_rxn.reactants:
                stoech_coeff = stoich_dict[met.id]
                # stoech_coeff < 0 indicates that the metabolite is a reactant
                if (stoech_coeff < 0) and (met not in all_sinks.keys()):
                    sink = Reaction("Sink_" + bio_rxn.id + "_" + met.id)
                    sink.name = "Sink_" + bio_rxn.name + "_" + met.name
                    # Subsystem specific to BBB sinks
                    sink.subsystem = "Demand"

                    # A sink is simply a reaction which consumes the BBB
                    sink.add_metabolites({met: -1})

                    # The sinks will be activated later (cf compute_lumps), one at a time
                    # sink.knock_out()

                    # The stoechiometric coefficients will be used to define the lower bound of the sink,
                    # thus it must be stored
                    all_sinks[met] = (sink.id, -stoech_coeff)
                    self._tfa_model.add_reactions([sink])

                # reactant already seen
                elif stoech_coeff < 0:
                    # The BBB has already been associated to a sink, so we simply increase the bound of the sink
                    all_sinks[met][1] -= stoech_coeff

        # Must be called before changing the reaction.thermo['computed'] values
        self._tfa_model.prepare()
        for ncrxn in self._rncore:
            ncrxn.thermo['computed'] = False
          
        return all_sinks

    def _generate_objective(self):
        """
        Generate and add the maximization objective : set as many activation variables as possible to 1
        When an activation variable is set to 1, the corresponding non-core reaction is deactivated
        """
        # Sum of binary variables to be maximized
        objective_sum = symbol_sum(list(self._activation_vars.values()))
        # Set the sum as the objective function
        self._tfa_model.objective = self._tfa_model.problem.Objective(objective_sum, direction='max')

    def compute_lumps(self, force_solve=False, method='OnePerBBB'):
        """
        For each BBB (reactant of the biomass reaction), add the corresponding sink to the model, then optimize and
        lump the result into one lumped reaction
        :param force_solve: Indicates whether the computations must continue when one lumping yields a status "infeasible"
        :return: The dict {BBB: lump} containing every lumped reactions, associated to their BBBs
        """

        # Must be called before optimization
        self._tfa_model.convert()
        # self._tfa_model.objective_direction = 'min'

        epsilon = self._tfa_model.solver.configuration.tolerances.feasibility

        the_method = disambiguate(method)
        print('Lumping method detected: {}'.format(the_method))

        # dict: {metabolite: lumped_reaction}
        lumps = {}

        self._tfa_model.objective_direction = 'max'

        sink_iter = tqdm(self._sinks.items(), desc = 'met')

        for met_BBB, (sink_id, stoech_coeff) in sink_iter:

            # Cute stuff
            sink_iter.set_description('met={}'.format(met_BBB.id[:10]))
            sink_iter.refresh()

            sink = self._tfa_model.reactions.get_by_id(sink_id)
            # Activate reaction by setting its lower bound
            prev_lb = sink.lower_bound
            min_prod = self._growth_rate * stoech_coeff
            sink.lower_bound = min_prod - epsilon

            if the_method == 'oneperbbb':
                this_lump = self._lump_one_per_bbb(met_BBB, sink, force_solve)
                lumped_reactions = [this_lump] if this_lump is not None else list()
            elif the_method.startswith('min+'):
                try:
                    p = int(the_method.replace('min+',''))
                except ValueError:
                    raise ValueError('Min+p method must have p as an integer')
                lumped_reactions = self._lump_min_plus_p(met_BBB, sink, p, force_solve)
            elif the_method.startswith('min'):
                lumped_reactions = self._lump_min_plus_p(met_BBB, sink, 0, force_solve)
            else:
                raise ValueError('Lumping method not recognized: {}. '
                                 'Valid methods are '
                                 'OnePerBBB, Min, Min+p, p natural integer'
                                 .format(the_method))


            if not lumped_reactions:
                continue

            lumps[met_BBB] = lumped_reactions

            # Deactivating reaction by setting both bounds to 0
            sink.lower_bound = prev_lb
            # sink.knock_out()

        self.lumps = lumps
        return lumps

    def _lump_one_per_bbb(self, met_BBB, sink, force_solve):
        """

        :param met_BBB:
        :param sink:
        :param force_solve:
        :return:
        """

        n_da = self._tfa_model.slim_optimize()

        try:
            # Timeout reached
            if self._tfa_model.solver.status == TIME_LIMIT:
                raise TimeoutExcept(self._tfa_model.solver.configuration.timeout)
            # Not optimal status -> infeasible
            elif self._tfa_model.solver.status != OPTIMAL:
                raise InfeasibleExcept(self._tfa_model.solver.status,
                                       self._tfa_model.solver.configuration.tolerances.feasibility)

        except (TimeoutExcept, InfeasibleExcept) as err:
            # If the user want to continue anyway, suits him
            if force_solve:
                # Raise a warning
                return None
            else:
                raise err

        # print('Produced {}'.format(sink.flux),
        #       'with {0:.0f} reactions deactivated'.format(n_da))

        lumped_reaction = self._build_lump(met_BBB, sink)

        return lumped_reaction


    def _lump_min_plus_p(self, met_BBB, sink, p, force_solve):
        """

        :param met_BBB:
        :param sink:
        :param force_solve:
        :return:
        """

        epsilon = self._tfa_model.solver.configuration.tolerances.integrality

        try:
            max_lumps =self._param_dict['max_lumps_per_BBB']
        except KeyError:
            # TODO: Put a warning
            max_lumps=10

        lumps = list()

        with self._tfa_model as model:
            activation_vars = model.get_variables_of_type(FluxKO)

            # Solve a first time, obtain minimal subnet
            model.slim_optimize()
            max_deactivated_rxns = model.objective.value

            # Add constraint forbidding subnets bigger than p
            expr = symbol_sum(activation_vars)

            # The lower bound is the max number of deactivated, minus p
            # Which allows activating the minimal number of reactions, plus p
            lb = max_deactivated_rxns - p
            model.add_constraint(kind=ForbiddenProfile,
                                 hook = model,
                                 id_ = 'MAX_DEACT_{}'.format(met_BBB.id),
                                 expr = expr,
                                 lb = lb,
                                 ub = max_deactivated_rxns,
                                 )

            n_deactivated_reactions = max_deactivated_rxns

            # While loop, break on infeasibility
            while len(lumps)<max_lumps:

                try:
                    this_lump = self._lump_one_per_bbb(met_BBB, sink, force_solve)
                except (InfeasibleExcept, TimeoutExcept) as e:
                    if force_solve:
                        pass
                    elif len(lumps) == 0:
                        # No solution AND no lump found
                        raise e

                if model.solver.status != OPTIMAL:
                    break
                elif this_lump is None:
                    # Since the solver is optimal, and we caught optim errors before,
                    # Then the BBB is simply produced in enough quantity by the core
                    break

                lumps.append(this_lump)

                # Add constraint forbidding the previous solution
                is_inactivated = [x for x in activation_vars
                               if abs(x.variable.primal-1) < 2*epsilon]

                expr = symbol_sum(is_inactivated)
                model.add_constraint(kind=ForbiddenProfile,
                                     hook = model,
                                     id_ = '{}_{}_{}'.format(met_BBB.id,
                                                             n_deactivated_reactions,
                                                             len(lumps)),
                                     expr = expr,
                                     lb = max_deactivated_rxns-p-1,
                                     ub = n_deactivated_reactions-1,
                                     )

        # TODO: Update of dynamic properties not handled yet
        # upon exiting context manager
        model.repair()
        return lumps


    def _build_lump(self, met_BBB, sink):
        """
        This function uses the current solution of self._tfa_model

        :param met_BBB:
        :param sink:
        :return:
        """

        epsilon_int = self._tfa_model.solver.configuration.tolerances.integrality
        epsilon_flux = self._tfa_model.solver.configuration.tolerances.feasibility

        sigma = sink.flux
        lump_dict = dict()

        for rxn in self._rncore:
            if self._activation_vars[rxn].variable.primal < epsilon_int:
                lump_dict[rxn] = rxn.flux / sigma
        # lumped_reaction1 = sum([rxn * (flux / sigma)
        #                       for rxn, flux in lump_dict.items()])

        if not lump_dict:
            # No need for lump
            self._tfa_model.logger.info('Metabolite {} is produced in enough '
                                        'quantity by core reactions'.format(met_BBB.id))
            return None

        lumped_reaction = sum_reactions(lump_dict,
                                        id_=sink.id.replace('Sink_', 'LUMP_'),
                                        epsilon = epsilon_flux)
        return lumped_reaction


def sum_reactions(rxn_dict, id_ = 'summed_reaction', epsilon = 1e-9):
    """
    Keys are reactions
    Values are their multiplicative coefficient
    """
    stoich = defaultdict(int)

    for rxn,flux in rxn_dict.items():
        for x, coeff in rxn.metabolites.items():
            stoich[x.id] += coeff * flux

    gpr = ') and ('.join(x.gene_reaction_rule for x in rxn_dict if x.gene_reaction_rule)

    gpr = ('(' + gpr + ')') if gpr else ''

    stoich = trim_epsilon_mets(stoich, epsilon=epsilon)

    new = Lump(id_ = id_,
               metabolites = stoich,
               subnetwork = {x.id:v for x,v in rxn_dict.items()},
               gene_reaction_rule=gpr)

    return new


# -*- coding: utf-8 -*-
"""
.. module:: pytfa
   :platform: Unix, Windows
   :synopsis: Thermodynamics-based Flux Analysis

.. moduleauthor:: pyTFA team

Thermodynamic cobra_model class and methods definition


"""

import re
from copy import deepcopy
from math import log

import pandas as pd
from cobra import Model

from ..core.model import LCSBModel
from . import std
from .metabolite import MetaboliteThermo
from .reaction import calcDGtpt_rhs, calcDGR_cues, get_debye_huckel_b
from .utils import (
    check_reaction_balance,
    check_transport_reaction,
    find_transported_mets,
    get_reaction_compartment,
)
from ..optim.constraints import (
    SimultaneousUse,
    NegativeDeltaG,
    BackwardDeltaGCoupling,
    ForwardDeltaGCoupling,
    BackwardDirectionCoupling,
    ForwardDirectionCoupling,
    ReactionConstraint,
    MetaboliteConstraint,
    DisplacementCoupling,
)
from ..optim.variables import (
    ThermoDisplacement,
    DeltaGstd,
    DeltaG,
    ForwardUseVariable,
    BackwardUseVariable,
    LogConcentration,
    ReactionVariable,
    MetaboliteVariable,
)
from ..utils import numerics
from ..utils.logger import get_bistream_logger

BIGM = numerics.BIGM
BIGM_THERMO = numerics.BIGM_THERMO
BIGM_DG = numerics.BIGM_DG
BIGM_P = numerics.BIGM_P
EPSILON = numerics.EPSILON
MAX_STOICH = 10


class ThermoModel(LCSBModel, Model):
    """
    A class representing a cobra_model with thermodynamics information

    """

    def __init__(self, thermo_data=None, model=Model(), name=None,
                 temperature=std.TEMPERATURE_0,
                 min_ph=std.MIN_PH,
                 max_ph=std.MAX_PH):

        """
        :param float temperature: the temperature (K) at which to perform the calculations
        :param dict thermo_data: The thermodynamic database
        :type temperature: float
        """

        LCSBModel.__init__(self, model, name)

        self.logger = get_bistream_logger('ME model' + str(self.name))

        self.TEMPERATURE = temperature
        self.thermo_data = thermo_data
        self.parent = model

        # CONSTANTS
        self.MAX_pH = max_ph
        self.MIN_pH = min_ph

        self._init_thermo()

        self.logger.info('# Model initialized with units {} and temperature {} K'  \
                    .format(self.thermo_unit, self.TEMPERATURE))

    def _init_thermo(self):

        self.thermo_unit = self.thermo_data['units']
        self.reaction_cues_data = self.thermo_data['cues']
        self.compounds_data = self.thermo_data['metabolites']
        self.Debye_Huckel_B = get_debye_huckel_b(self.TEMPERATURE)

        self.logger = get_bistream_logger('thermomodel_' + str(self.name))

        # Compute internal values to adapt the the thermo_unit provided
        if self.thermo_unit == "kJ/mol":
            self.GAS_CONSTANT = 8.314472 / 1000  # kJ/(K mol)
            self.Adjustment = 1
        else:
            self.GAS_CONSTANT = 1.9858775 / 1000  # Kcal/(K mol)
            self.Adjustment = 4.184

        self.RT = self.GAS_CONSTANT * self.TEMPERATURE

    def normalize_reactions(self):
        """
        Find reactions with important stoichiometry and normalizes them
        :return:
        """

        self.logger.info("# Model normalization")

        for this_reaction in self.reactions:
            metabolites = this_reaction.metabolites
            max_stoichiometry = max(metabolites.values())
            if max_stoichiometry > MAX_STOICH:
                new_metabolites = {
                    k: -v + v / max_stoichiometry
                    for k, v in metabolites.items()
                }
                this_reaction.add_metabolites(new_metabolites)
            else:
                continue

    def _prepare_metabolite(self, met):
        """

        :param met:
        :return:
        """

        # Get the data about the compartment of the enzyme
        if not met.compartment in self.compartments:
            raise Exception(
                "Compartment not found in cobra_model : " + met.compartment
            )

        CompartmentpH = self.compartments[met.compartment]["pH"]
        CompartmentionicStr = self.compartments[met.compartment]["ionicStr"]

        # Which index of the reaction DB do you correspond to ?
        if not "seed_id" in met.annotation:
            # raise Exception("seed_id missing for " + met.name)
            self.logger.debug(
                "Metabolite {} ({}) has no seed_id".format(met.id, met.name)
            )
            metData = None
        elif not met.annotation["seed_id"] in self.compounds_data:
            self.logger.debug(
                "Metabolite {} ({}) not present in thermoDB".format(
                    met.annotation["seed_id"], met.name
                )
            )
            metData = None
        else:
            metData = self.compounds_data[met.annotation["seed_id"]]
            # Override the formula
            met.formula = metData["formula"]

        met.thermo = MetaboliteThermo(
            metData,
            CompartmentpH,
            CompartmentionicStr,
            self.TEMPERATURE,
            self.MIN_pH,
            self.MAX_pH,
            self.Debye_Huckel_B,
            self.thermo_unit,
        )

    def _prepare_reaction(self, reaction, null_error_override=2):
        """

        :param reaction:
        :param null_error_override: overrides DeltaG when it is 0 to
                                    allow flexibility. 2kcal/mol is standard in
                                    estimation frameworks like GCM.
        :return:
        """
        DeltaGrxn = 0
        DeltaGRerr = 0
        proton_of = self._proton_of

        # identifying the reactants
        if len(reaction.metabolites) == 1:
            NotDrain = False
        else:
            NotDrain = True

        # Initialize a dictionnary where we will put our data - FIXME : Create a thermo object ?
        reaction.thermo = {"isTrans": False}

        # also check if rxn and enzyme compartments match
        reaction.compartment = get_reaction_compartment(reaction)

        # Make sure the reaction is balanced...

        balanceResult = check_reaction_balance(
            reaction,
            (
                proton_of[reaction.compartment]
                if reaction.compartment in proton_of
                else None
            ),
        )

        # Also test if this is a transport reaction
        reaction.thermo["isTrans"] = check_transport_reaction(reaction)
        # Make sure we have correct thermo values for each metabolites
        correctThermoValues = True

        for met in reaction.metabolites:
            if met.thermo.deltaGf_std > 0.9 * BIGM_DG:
                correctThermoValues = False
                break

        if (
            not NotDrain
            or not correctThermoValues
            or len(reaction.metabolites) >= 100
            or balanceResult in ["missing atoms", "drain flux"]
        ):

            self.logger.debug(
                "{} : thermo constraint NOT created".format(reaction.id)
            )
            reaction.thermo["computed"] = False
            reaction.thermo["deltaGR"] = BIGM_DG
            reaction.thermo["deltaGRerr"] = BIGM_DG

        else:
            self.logger.debug(
                "{} : thermo constraint created".format(reaction.id)
            )
            reaction.thermo["computed"] = True

            if reaction.thermo["isTrans"]:
                (rhs, breakdown) = calcDGtpt_rhs(
                    reaction, self.compartments, self.thermo_unit
                )

                reaction.thermo["deltaGR"] = rhs

                reaction.thermo["deltaGrxn"] = breakdown["sum_deltaGFis"]
            else:
                for met in reaction.metabolites:
                    if met.formula != "H" or (
                        "seed_id" in met.annotation
                        # That's H+
                        and met.annotation["seed_id"] != "cpd00067"
                    ):
                        DeltaGrxn += (
                            reaction.metabolites[met] * met.thermo.deltaGf_tr
                        )
                        DeltaGRerr += abs(
                            reaction.metabolites[met] * met.thermo.deltaGf_err
                        )

                reaction.thermo["deltaGR"] = DeltaGrxn

            (tmp1, DeltaGRerr, tmp2, tmp3) = calcDGR_cues(
                reaction, self.reaction_cues_data
            )

            if DeltaGRerr == 0 and null_error_override:
                DeltaGRerr = null_error_override  # default value for DeltaGRerr

            reaction.thermo["deltaGRerr"] = DeltaGRerr

    def prepare(self, null_error_override=2):
        """ Prepares a COBRA toolbox cobra_model for TFBA analysis by doing the following:

           1. checks if a reaction is a transport reaction
           2. checks the ReactionDB for Gibbs energies of formation of metabolites
           3. computes the Gibbs energies of reactions

        :param null_error_override: overrides DeltaG when it is 0 to
                                    allow flexibility. 2kcal/mol is standard in
                                    estimation frameworks like GCM.

        """

        self.logger.info("# Model preparation starting...")

        # Number of reactions
        num_rxns = len(self.reactions)

        # Number of metabolites
        num_mets = len(self.metabolites)

        for i in range(num_mets):
            met = self.metabolites[i]
            self._prepare_metabolite(met)

        # And now, reactions !

        self.logger.debug("computing reaction thermodynamic data")

        # Look for the proton enzyme...
        proton = {}
        for i in range(num_mets):
            if self.metabolites[i].formula == "H" or (
                "seed_id" in self.metabolites[i].annotation
                and self.metabolites[i].annotation["seed_id"] == "cpd00067"
            ):
                proton[self.metabolites[i].compartment] = self.metabolites[i]

        if len(proton) == 0:
            raise Exception("Cannot find proton")
        else:
            self._proton_of = proton

        # Iterate over each reaction
        for i in range(num_rxns):
            reaction = self.reactions[i]
            self._prepare_reaction(reaction, null_error_override)

        self.logger.info("# Model preparation done.")

    def _convert_metabolite(self, met, add_potentials, verbose):
        """
        Given a enzyme, proceeds to create the necessary variables and
        constraints for thermodynamics-based modeling

        :param met:
        :return:
        """

        if add_potentials:
            P_lb = -BIGM_P  # kcal/mol
            P_ub = BIGM_P  # kcal/mol

        # exclude protons and water and those without deltaGF
        # P_met: P_met - RT*LC_met = DGF_met
        metformula = met.formula
        metDeltaGF = met.thermo.deltaGf_tr
        metComp = met.compartment
        metLConc_lb = log(self.compartments[metComp]["c_min"])
        metLConc_ub = log(self.compartments[metComp]["c_max"])
        Comp_pH = self.compartments[metComp]["pH"]
        LC = None

        if metformula == "H2O":
            LC = self.add_variable(LogConcentration, met, lb=0, ub=0)

        elif metformula == "H":
            LC = self.add_variable(
                LogConcentration,
                met,
                lb=log(10 ** -Comp_pH),
                ub=log(10 ** -Comp_pH),
            )

        elif (
            "seed_id" in met.annotation
            and met.annotation["seed_id"] == "cpd11416"
        ):
            # we do not create the thermo variables for biomass enzyme
            pass

        elif metDeltaGF < 10 ** 6:
            if verbose:
                self.logger.debug(
                    "generating thermo variables for {}".format(met.id)
                )
            LC = self.add_variable(
                LogConcentration, met, lb=metLConc_lb, ub=metLConc_ub
            )

            if add_potentials:
                P = self.add_variable("P_" + met.id, P_lb, P_ub)
                self.P_vars[met] = P
                # Formulate the constraint
                expr = P - self.RT * LC
                self.add_constraint(
                    LogConcentration,
                    "P_" + met.id,
                    expr,
                    metDeltaGF,
                    metDeltaGF,
                )

        else:
            self.logger.debug(
                "NOT generating thermo variables for {}".format(met.id)
            )

        if LC != None:
            # Register the variable to find it more easily
            self.LC_vars[met] = LC

    def _convert_reaction(self, rxn, add_potentials, add_displacement, verbose):
        """

        :param rxn:
        :param add_potentials:
        :param add_displacement:
        :param verbose:
        :return:
        """

        RT = self.RT

        DGR_lb = -BIGM_THERMO  # kcal/mol
        DGR_ub = BIGM_THERMO  # kcal/mol

        epsilon = self.solver.configuration.tolerances.feasibility

        # Is it a water transport reaction ?
        H2OtRxns = False
        if rxn.thermo["isTrans"] and len(rxn.reactants) == 1:
            try:
                if rxn.reactants[0].annotation["seed_id"] == "cpd00001":
                    H2OtRxns = True
            except KeyError:
                pass

        # Is it a drain reaction ?
        NotDrain = len(rxn.metabolites) > 1

        # if the reaction is flagged with rxnThermo, and it's not a H2O
        # transport, we will add thermodynamic constraints
        if rxn.thermo["computed"] and not H2OtRxns:
            if verbose:
                self.logger.debug(
                    "generating thermo constraint for {}".format(rxn.id)
                )

            # add the delta G as a variable
            DGR = self.add_variable(DeltaG, rxn, lb=DGR_lb, ub=DGR_ub)

            # add the delta G naught as a variable
            RxnDGerror = rxn.thermo["deltaGRerr"]
            DGoR = self.add_variable(
                DeltaGstd,
                rxn,
                lb=rxn.thermo["deltaGR"] - RxnDGerror,
                ub=rxn.thermo["deltaGR"] + RxnDGerror,
            )

            # Initialization of indices and coefficients for all possible
            # scenaria:
            LC_TransMet = 0
            LC_ChemMet = 0
            P_expr = 0

            if rxn.thermo["isTrans"]:
                # calculate the DG component associated to transport of the
                # enzyme. This will be added to the constraint on the Right
                # Hand Side (RHS)

                transportedMets = find_transported_mets(rxn)

                # Chemical coefficient, it is the enzyme's coefficient...
                # + transport coeff for reactants
                # - transport coeff for products
                chem_stoich = rxn.metabolites.copy()

                # Adding the terms for the transport part
                for seed_id, trans in transportedMets.items():
                    for type_ in ["reactant", "product"]:
                        if trans[type_].formula != "H":
                            LC_TransMet += (
                                self.LC_vars[trans[type_]]
                                * RT
                                * trans["coeff"]
                                * (-1 if type_ == "reactant" else 1)
                            )

                        chem_stoich[trans[type_]] += trans["coeff"] * (
                            1 if type_ == "reactant" else -1
                        )

                # Also add the chemical reaction part
                chem_stoich = {
                    met: val for met, val in chem_stoich.items() if val != 0
                }

                for met in chem_stoich:
                    metFormula = met.formula
                    if metFormula not in ["H", "H2O"]:
                        LC_ChemMet += self.LC_vars[met] * RT * chem_stoich[met]

            else:
                # if it is just a regular chemical reaction
                if add_potentials:
                    RHS_DG = 0

                    for met in rxn.metabolites:
                        metformula = met.formula
                        metDeltaGFtr = met.thermo.deltaGf_tr
                        if metformula == "H2O":
                            RHS_DG = (
                                RHS_DG + rxn.metabolites[met] * metDeltaGFtr
                            )
                        elif metformula != "H":
                            P_expr += self.P_vars[met] * rxn.metabolites[met]
                else:
                    # RxnDGnaught on the right hand side
                    RHS_DG = rxn.thermo["deltaGR"]

                    for met in rxn.metabolites:
                        metformula = met.formula
                        if metformula not in ["H", "H2O"]:
                            # we use the LC here as we already accounted for the
                            # changes in deltaGFs in the RHS term
                            LC_ChemMet += (
                                self.LC_vars[met] * RT * rxn.metabolites[met]
                            )

            # G: - DGR_rxn + DGoRerr_Rxn
            #   + RT * StoichCoefProd1 * LC_prod1
            #   + RT * StoichCoefProd2 * LC_prod2
            #   + RT * StoichCoefSub1 * LC_subs1
            #   + RT * StoichCoefSub2 * LC_subs2
            #   - ...
            #   = 0

            # Formulate the constraint
            CLHS = DGoR - DGR + LC_TransMet + LC_ChemMet
            self.add_constraint(NegativeDeltaG, rxn, CLHS, lb=0, ub=0)

            if add_displacement:
                lngamma = self.add_variable(
                    ThermoDisplacement, rxn, lb=-BIGM_P, ub=BIGM_P
                )

                # ln(Gamma) = +DGR/RT (DGR < 0 , rxn is forward, ln(Gamma) < 0d
                expr = lngamma - 1 / RT * DGR
                self.add_constraint(DisplacementCoupling, rxn, expr, lb=0, ub=0)

            # Create the use variables constraints and connect them to the
            # deltaG if the reaction has thermo constraints
            # FU_rxn: 1000 FU_rxn + DGR_rxn < 1000 - epsilon
            FU_rxn = self.add_variable(ForwardUseVariable, rxn)

            CLHS = DGR + FU_rxn * BIGM_THERMO
            self.add_constraint(
                ForwardDeltaGCoupling, rxn, CLHS, ub=BIGM_THERMO - epsilon
            )

            # BU_rxn: 1000 BU_rxn - DGR_rxn < 1000 - epsilon
            BU_rxn = self.add_variable(BackwardUseVariable, rxn)

            CLHS = BU_rxn * BIGM_THERMO - DGR
            self.add_constraint(
                BackwardDeltaGCoupling, rxn, CLHS, ub=BIGM_THERMO - epsilon
            )

        else:
            if not NotDrain:
                self.logger.debug(
                    "generating only use constraints for drain reaction"
                    + rxn.id
                )
            else:
                self.logger.debug(
                    "generating only use constraints for reaction" + rxn.id
                )

            FU_rxn = self.add_variable(ForwardUseVariable, rxn)
            BU_rxn = self.add_variable(BackwardUseVariable, rxn)

        # create the prevent simultaneous use constraints
        # SU_rxn: FU_rxn + BU_rxn <= 1
        CLHS = FU_rxn + BU_rxn
        self.add_constraint(SimultaneousUse, rxn, CLHS, ub=1)

        # create constraints that control fluxes with their use variables
        # UF_rxn: F_rxn - M FU_rxn < 0
        F_rxn = rxn.forward_variable
        CLHS = F_rxn - FU_rxn * BIGM
        self.add_constraint(ForwardDirectionCoupling, rxn, CLHS, ub=0)

        # UR_rxn: R_rxn - M RU_rxn < 0
        R_rxn = rxn.reverse_variable
        CLHS = R_rxn - BU_rxn * BIGM
        self.add_constraint(BackwardDirectionCoupling, rxn, CLHS, ub=0)

    def convert(
        self, add_potentials=False, add_displacement=False, verbose=True
    ):
        """ Converts a cobra_model into a tFBA ready cobra_model by adding the
        thermodynamic constraints required

        .. warning::
            This function requires you to have already called
            :func:`~.pytfa.ThermoModel.prepare`, otherwise it will raise an Exception !

        """

        self.logger.info("# Model conversion starting...")

        ###########################################
        # CONSTANTS & PARAMETERS for tFBA problem #
        ###########################################

        # value for the bigM in big M constraints such as:
        # UF_rxn: F_rxn - M*FU_rxn < 0
        bigM = BIGM
        # Check each reactions' bounds
        for reaction in self.reactions:
            if (
                reaction.lower_bound < -bigM - EPSILON
                or reaction.upper_bound > bigM + EPSILON
            ):
                raise Exception("flux bounds too wide or big M not big enough")
            if reaction.lower_bound < -bigM:
                reaction.lower_bound = -bigM
            if reaction.upper_bound > bigM:
                reaction.upper_bound = bigM

        ###################
        # INPUTS & CHECKS #
        ###################

        # check if cobra_model reactions has been checked if they are transport reactions
        try:
            for reaction in self.reactions:
                if not "isTrans" in reaction.thermo:
                    reaction.thermo["isTrans"] = check_transport_reaction(
                        reaction
                    )
        except:
            raise Exception(
                "Reaction thermo data missing. "
                + "Please run ThermoModel.prepare()"
            )

        # FIXME Use generalized rule (ext to the function)
        # formatting the enzyme and reaction names to remove brackets
        replacements = {"_": re.compile(r"[\[\(]"), "": re.compile(r"[\]\)]")}
        for items in [self.metabolites, self.reactions]:
            for item in items:
                for rep in replacements:
                    item.name = re.sub(replacements[rep], rep, item.name)

        self.LC_vars = {}
        self.P_vars = {}

        for met in self.metabolites:
            self._convert_metabolite(met, add_potentials, verbose)

        ## For each reaction...
        for rxn in self.reactions:
            self._convert_reaction(
                rxn, add_potentials, add_displacement, verbose
            )

        # CONSISTENCY CHECKS

        # Creating the objective
        if len(self.objective.variables) == 0:
            self.logger.warning("Objective not found")

        self.logger.info("# Model conversion done.")
        self.logger.info("# Updating cobra_model variables...")
        self.repair()
        self.logger.info("# cobra_model variables are up-to-date")

    def print_info(self, specific=False):
        """
        Print information and counts for the cobra_model
        :return:
        """
        if not specific:
            LCSBModel.print_info(self)

        n_metabolites = len(self.metabolites)
        n_reactions = len(self.reactions)
        n_metabolites_thermo = len(
            # [
            #     x
            #     for x in self.metabolites
            #     if hasattr(x, "thermo") and x.thermo["id"]
            # ]
            self._var_kinds[LogConcentration.__name__]
        )
        n_reactions_thermo = len(
            # [
            #     x
            #     for x in self.reactions
            #     if x.id is not None
            #     and hasattr(x, "thermo")
            #     and x.thermo["computed"]
            # ]
            self._cons_kinds[ForwardDeltaGCoupling.__name__]
        )

        info = pd.DataFrame(columns=["value"])
        info.loc["num metabolites(thermo)"] = n_metabolites_thermo
        info.loc["num reactions(thermo)"] = n_reactions_thermo
        info.loc["pct metabolites(thermo)"] = (
            n_metabolites_thermo / n_metabolites * 100
        )
        info.loc["pct reactions(thermo)"] = (
            n_reactions_thermo / n_reactions * 100
        )
        info.index.name = "key"

        print(info)

    def __deepcopy__(self, memo):
        """

        :param memo:
        :return:
        """

        return self.copy()

    def copy(self):

        from ..io.dict import model_from_dict, model_to_dict
        from ..optim.utils import copy_solver_configuration

        dictmodel = model_to_dict(self)
        new = model_from_dict(dictmodel)

        copy_solver_configuration(self, new)

        return new
